"""
Request timeout in seconds for Looker API
"""

LOOKER_ENV_PREFIX = "CASTOR_LOOKER_"

DEFAULT_LOOKER_TIMEOUT_SECOND = 120
"""
Number of items per page when requesting Looker API
"""
DEFAULT_LOOKER_PAGE_SIZE = 500
"""
Maximum concurrent threads to run when fetching
"""
DEFAULT_LOOKER_THREAD_POOL_SIZE = 20
MIN_THREAD_POOL_SIZE = 1
MAX_THREAD_POOL_SIZE = 200



import logging
from functools import partial
from typing import Iterator, Optional

import requests

from ...utils import (
    APIClient,
    BearerAuth,
    SafeMode,
    build_url,
    fetch_all_pages,
    handle_response,
    retry,
    safe_mode,
)
from ..abstract import TimeFilter
from .credentials import DatabricksCredentials
from .endpoints import DatabricksEndpointFactory
from .format import DatabricksFormatter, TagMapping
from .pagination import DATABRICKS_PAGE_SIZE, DatabricksPagination
from .types import TablesColumns
from .utils import hourly_time_filters

logger = logging.getLogger(__name__)

_DATABRICKS_CLIENT_TIMEOUT_S = 90
_MAX_NUMBER_OF_LINEAGE_ERRORS = 1000
_MAX_NUMBER_OF_QUERY_ERRORS = 1000
_RETRY_ATTEMPTS = 3
_RETRY_BASE_MS = 1000
_RETRY_EXCEPTIONS = [
    requests.exceptions.ConnectTimeout,
]
_WORKSPACE_ID_HEADER = "X-Databricks-Org-Id"

safe_lineage_params = SafeMode((BaseException,), _MAX_NUMBER_OF_LINEAGE_ERRORS)
safe_query_params = SafeMode((BaseException,), _MAX_NUMBER_OF_QUERY_ERRORS)


class DatabricksAuth(BearerAuth):
    def __init__(self, credentials: DatabricksCredentials):
        self.token = credentials.token

    def fetch_token(self) -> Optional[str]:
        return self.token


class DatabricksAPIClient(APIClient):
    """Databricks Client"""

    def __init__(
        self,
        credentials: DatabricksCredentials,
        db_allowed: Optional[set[str]] = None,
        db_blocked: Optional[set[str]] = None,
    ):
        auth = DatabricksAuth(credentials)
        super().__init__(
            host=credentials.host,
            auth=auth,
            timeout=_DATABRICKS_CLIENT_TIMEOUT_S,
        )
        self._http_path = credentials.http_path
        self._db_allowed = db_allowed
        self._db_blocked = db_blocked

        self.formatter = DatabricksFormatter()

    def _keep_catalog(self, catalog: str) -> bool:
        """
        Helper function to determine if we should keep the Databricks catalog
        which is a CastorDoc database
        """
        if self._db_allowed and catalog not in self._db_allowed:
            return False
        if self._db_blocked and catalog in self._db_blocked:
            return False
        return True

    def databases(self) -> list[dict]:
        content = self._get(DatabricksEndpointFactory.databases())
        _databases = self.formatter.format_database(content.get("catalogs", []))
        filtered_databases = [
            d for d in _databases if self._keep_catalog(d["database_name"])
        ]
        logger.info(
            f"Available databases: {[d['database_name'] for d in filtered_databases]}"
        )
        return filtered_databases

    def _schemas_of_database(self, database: dict) -> list[dict]:
        payload = {"catalog_name": database["database_name"]}
        content = self._get(DatabricksEndpointFactory.schemas(), params=payload)
        schemas = content.get("schemas", [])
        return self.formatter.format_schema(schemas, database)

    def schemas(self, databases: list[dict]) -> list[dict]:
        """
        Get the databricks schemas (also sometimes called databases)
        (which correspond to the schemas in Castor)
        leveraging the unity catalog API
        """
        return [
            schema
            for database in databases
            for schema in self._schemas_of_database(database)
        ]

    def tables_columns_of_schema(
        self,
        schema: dict,
        table_tags: TagMapping,
        column_tags: TagMapping,
    ) -> TablesColumns:
        payload = {
            "catalog_name": schema["database_id"],
            "schema_name": schema["schema_name"],
        }
        response = self._call(
            method="GET",
            endpoint=DatabricksEndpointFactory.tables(),
            params=payload,
        )
        workspace_id = response.headers[_WORKSPACE_ID_HEADER]
        content = handle_response(response)
        host = build_url(self._host, endpoint="")
        return self.formatter.format_table_column(
            raw_tables=content.get("tables", []),
            schema=schema,
            host=host,
            workspace_id=workspace_id,
            table_tags=table_tags,
            column_tags=column_tags,
        )

    @safe_mode(safe_query_params, lambda: [])
    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_ATTEMPTS,
        base_ms=_RETRY_BASE_MS,
    )
    def _queries(
        self,
        filter_: dict,
    ) -> Iterator[dict]:
        """
        Callback to scroll the queries api
        https://docs.databricks.com/api/workspace/queryhistory/list
        max_results: Limit the number of results returned in one page.
            The default is 100. (both on our side and Databricks')
        """
        payload = {**filter_, "max_results": DATABRICKS_PAGE_SIZE}
        request = partial(
            self._get,
            endpoint=DatabricksEndpointFactory.queries(),
            data=payload,
        )
        queries = fetch_all_pages(request, DatabricksPagination)
        return queries

    def queries(self, time_filter: Optional[TimeFilter] = None) -> list[dict]:
        """get all queries, hour per hour"""
        time_range_filters = hourly_time_filters(time_filter)
        raw_queries = []
        for _filter in time_range_filters:
            logger.info(f"Fetching queries for time filter {_filter}")
            hourly = self._queries(_filter)
            raw_queries.extend(hourly)
        return self.formatter.format_query(raw_queries)

    def users(self) -> list[dict]:
        """
        retrieve user from api
        """
        content = self._get(DatabricksEndpointFactory.users())
        return self.formatter.format_user(content.get("Resources", []))

    def _view_ddl_per_schema(self, schema: dict) -> list[dict]:
        payload = {
            "catalog_name": schema["database_id"],
            "schema_name": schema["schema_name"],
            "omit_columns": True,
        }
        content = self._get(DatabricksEndpointFactory.tables(), params=payload)
        return self.formatter.format_view_ddl(content.get("tables", []), schema)

    def view_ddl(self, schemas: list[dict]) -> list[dict]:
        """retrieve view ddl"""
        view_ddl: list[dict] = []
        for schema in schemas:
            v_to_add = self._view_ddl_per_schema(schema)
            view_ddl.extend(v_to_add)
        return view_ddl



from ...types import ExternalAsset


class TableauAsset(ExternalAsset):
    """
    Tableau assets
    """

    COLUMN = "columns"
    DASHBOARD = "dashboards"
    DATASOURCE = "datasources"
    FIELD = "fields"
    METRIC = "metrics"
    METRIC_DEFINITION = "metrics_definitions"
    PROJECT = "projects"
    SHEET = "sheets"
    SUBSCRIPTION = "subscriptions"
    TABLE = "tables"
    USAGE = "usage"
    USER = "users"
    WORKBOOK = "workbooks"


# assets that are only available for clients using Tableau Pulse
TABLEAU_PULSE_ASSETS = (
    TableauAsset.METRIC,
    TableauAsset.METRIC_DEFINITION,
    TableauAsset.SUBSCRIPTION,
)



from .api_client import DatabricksAPIClient


class MockDatabricksClient(DatabricksAPIClient):
    def __init__(self):
        self._db_allowed = ["prd", "staging"]
        self._db_blocked = ["dev"]


def test_DatabricksAPIClient__keep_catalog():
    client = MockDatabricksClient()
    assert client._keep_catalog("prd")
    assert client._keep_catalog("staging")
    assert not client._keep_catalog("dev")
    assert not client._keep_catalog("something_unknown")



from typing import Optional

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_SCOPE = "https://analysis.windows.net/powerbi/api/.default"
POWERBI_ENV_PREFIX = "CASTOR_POWERBI_"

CLIENT_APP_BASE = "https://login.microsoftonline.com"
REST_API_BASE_PATH = "https://api.powerbi.com/v1.0/myorg"


class PowerbiCredentials(BaseSettings):
    """Class to handle PowerBI rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=POWERBI_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    client_id: str
    tenant_id: str
    secret: str = Field(repr=False)
    api_base: str = REST_API_BASE_PATH
    login_url: str = CLIENT_APP_BASE
    scopes: list[str] = [DEFAULT_SCOPE]

    @field_validator("scopes", mode="before")
    @classmethod
    def _check_scopes(cls, scopes: Optional[list[str]]) -> list[str]:
        return scopes if scopes is not None else [DEFAULT_SCOPE]

    @field_validator("login_url", mode="before")
    @classmethod
    def _check_login_url(cls, login_url: Optional[str]) -> str:
        return login_url if login_url is not None else CLIENT_APP_BASE

    @field_validator("api_base", mode="before")
    @classmethod
    def _check_api_base(cls, api_base: Optional[str]) -> str:
        return api_base if api_base is not None else REST_API_BASE_PATH



from collections.abc import Iterable
from datetime import date
from typing import Optional

from ...utils import at_midnight
from ..abstract import TimeFilter

_DEFAULT_HOUR_MIN = 0
_DEFAULT_HOUR_MAX = 23
_NUM_HOURS_IN_A_DAY = 24


def _day_hour_to_epoch_ms(day: date, hour: int) -> int:
    return int(at_midnight(day).timestamp() * 1000) + (hour * 3600 * 1000)


def build_path(
    row: dict,
    keys: list[str],
) -> str:
    """
    format an asset's path:
    - picks the given keys from dict
    - join keys with a dot "."
    """
    key_values = [row[key] for key in keys]
    return ".".join(key_values)


def tag_label(row: dict) -> str:
    """
    format the tag's label:
    - {key:value} when the value is not empty
    - {key} otherwise
    """
    tag_name = row["tag_name"]
    tag_value = row["tag_value"]
    if not tag_value:
        return tag_name
    return f"{tag_name}:{tag_value}"


def _time_filter_payload(start_time_ms: int, end_time_ms: int) -> dict:
    return {
        "filter_by": {
            "query_start_time_range": {
                "end_time_ms": end_time_ms,
                "start_time_ms": start_time_ms,
            }
        }
    }


def hourly_time_filters(time_filter: Optional[TimeFilter]) -> Iterable[dict]:
    """time filters to retrieve Databricks' queries: 1h duration each"""
    # define an explicit time window
    if not time_filter:
        time_filter = TimeFilter.default()

    assert time_filter  # for mypy

    hour_min = time_filter.hour_min
    hour_max = time_filter.hour_max
    day = time_filter.day
    if hour_min is None or hour_max is None:  # fallback to an entire day
        hour_min, hour_max = _DEFAULT_HOUR_MIN, _DEFAULT_HOUR_MAX

    for index in range(hour_min, min(hour_max + 1, _NUM_HOURS_IN_A_DAY)):
        start_time_ms = _day_hour_to_epoch_ms(day, index)
        end_time_ms = _day_hour_to_epoch_ms(day, index + 1)
        yield _time_filter_payload(start_time_ms, end_time_ms)



import io
from http import HTTPStatus

import pytest
from requests import HTTPError, Response

from .safe_request import (
    RequestSafeMode,
    handle_response,
)


def mock_response(status_code: int):
    response = Response()
    response.status_code = status_code
    response.raw = io.BytesIO(b'[{"data": "working"}]')
    return response


def test_http_error_with_no_safe_mode():
    with pytest.raises(HTTPError):
        handle_response(mock_response(HTTPStatus.FORBIDDEN))


def test_http_error_with_no_status_code():
    safe_params = RequestSafeMode(2)  # Caught

    with pytest.raises(HTTPError):
        handle_response(mock_response(HTTPStatus.FORBIDDEN), safe_params)


def test_http_error_with_status_code():
    safe_params = RequestSafeMode(2, (HTTPStatus.FORBIDDEN,))  # Caught

    def call():
        return handle_response(mock_response(HTTPStatus.FORBIDDEN), safe_params)

    assert call() == {}
    assert call() == {}

    with pytest.raises(HTTPError):
        call()


def test_http_error_with_multiple_status_code():
    safe_params = RequestSafeMode(
        2, (HTTPStatus.NOT_FOUND, HTTPStatus.FORBIDDEN)
    )  # Caught

    def call():
        return handle_response(mock_response(HTTPStatus.FORBIDDEN), safe_params)

    def call_2():
        return handle_response(mock_response(HTTPStatus.NOT_FOUND), safe_params)

    assert call() == {}
    assert call_2() == {}
    with pytest.raises(HTTPError):  # 3 failed calls > retries
        call()


def test_http_error_with_wrong_status_code():
    safe_params = RequestSafeMode(2, (HTTPStatus.NOT_FOUND,))  # Wrong Status

    def call():
        handle_response(mock_response(HTTPStatus.BAD_REQUEST), safe_params)

    with pytest.raises(HTTPError):
        call()


def test_http_error_with_return():
    safe_params = RequestSafeMode(2, (HTTPStatus.NOT_FOUND,))  # Wrong Status

    def call():
        return handle_response(mock_response(HTTPStatus.OK), safe_params)

    assert call() == [{"data": "working"}]



from typing import Optional

from ...utils import PaginationModel
from ...utils.salesforce.pagination import LIMIT_RECORDS_PER_PAGE
from .soql import (
    SOBJECTS_QUERY_TPL,
)

# Implicit (hard-coded in Salesforce) limitation when using SOQL of 2,000 rows
FIRST_START_DURABLE_ID = "0000"


def format_sobject_query(
    start_durable_id: str = FIRST_START_DURABLE_ID,
) -> str:
    return SOBJECTS_QUERY_TPL.format(
        start_durable_id=start_durable_id,
        limit=LIMIT_RECORDS_PER_PAGE,
    )


class SalesforceSQLPagination(PaginationModel):
    records: list

    def next_page_payload(self) -> Optional[dict]:
        start_durable_id = self.records[-1]["DurableId"]
        query = format_sobject_query(start_durable_id)
        return {"q": query}

    def is_last(self) -> bool:
        return len(self.records) < LIMIT_RECORDS_PER_PAGE

    def page_results(self) -> list:
        return self.records



from typing import Iterator, Optional

from google.oauth2.service_account import Credentials
from googleapiclient import discovery  # type: ignore

from ....utils import (
    at_midnight,
    current_date,
    fetch_all_pages,
    format_rfc_3339_date,
    past_date,
)
from .credentials import LookerStudioCredentials
from .pagination import LookerStudioPagination

USER_EMAIL_FIELD = "primaryEmail"


class AdminSDKClient:
    """
    Client to call the Report API and Directory API.
    The service account must impersonate and admin account.
    """

    def __init__(self, credentials: LookerStudioCredentials):
        self._credentials = Credentials.from_service_account_info(
            credentials.model_dump(),
            scopes=credentials.scopes,
            subject=credentials.admin_email,  # impersonates an admin
        )
        self.directory_api = discovery.build(
            "admin", "directory_v1", credentials=self._credentials
        )
        self.report_api = discovery.build(
            "admin", "reports_v1", credentials=self._credentials
        )

    def list_users(self) -> Iterator[dict]:
        """
        Lists all users in the domain; only the primaryEmail field is selected.
        Note:
        * `my_customer` is an alias to represent the account's `customerId`
        * `domain_public` allows non-admins to list users. This is technically
           not necessary here because an admin account is impersonated, but it
           avoids tapping into unnecessary data & serves for future reference.
        See
            https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.users.html#list
            https://developers.google.com/admin-sdk/directory/reference/rest/v1/users/list
            https://developers.google.com/admin-sdk/directory/v1/guides/manage-users#retrieve_users_non_admin
            https://stackoverflow.com/a/71083443/14448410
        """

        def _users(pagination_params: Optional[dict] = None) -> dict:
            parameters = {
                "viewType": "domain_public",
                "customer": "my_customer",
                "fields": f"users({USER_EMAIL_FIELD}), nextPageToken",
                **(pagination_params or {}),
            }

            return self.directory_api.users().list(**parameters).execute()

        yield from fetch_all_pages(_users, LookerStudioPagination)

    def list_view_events(self) -> Iterator[dict]:
        """
        Lists all Data Studio View events of the past day.
        See
            https://googleapis.github.io/google-api-python-client/docs/dyn/admin_reports_v1.activities.html
            https://developers.google.com/admin-sdk/reports/reference/rest/v1/activities/list
            https://developers.google.com/admin-sdk/reports/v1/appendix/activity/data-studio#VIEW
        """

        def _activity(pagination_params: Optional[dict] = None) -> dict:
            yesterday = format_rfc_3339_date(at_midnight(past_date(1)))
            today = format_rfc_3339_date(at_midnight(current_date()))

            parameters = {
                "userKey": "all",
                "applicationName": "data_studio",
                "eventName": "VIEW",
                "startTime": yesterday,
                "endTime": today,
                **(pagination_params or {}),
            }

            return self.report_api.activities().list(**parameters).execute()

        yield from fetch_all_pages(_activity, LookerStudioPagination)



from typing import Callable, Optional

from dateutil.parser import parse

from ..utils import string_to_tuple
from .enums import DataType, Issue

_CONVERTERS: dict[DataType, Callable] = {
    DataType.DATETIME: parse,
    DataType.FLOAT: float,
    DataType.INTEGER: int,
    DataType.STRING: str,
    DataType.LIST: string_to_tuple,
}


class ColumnChecker:
    """
    Provides rules and methods to validate column content (FileChecker).

    Handles the necessary context for validation:
    - Set of occurrences (to check unicity)
    - Set of foreign values (to check FK)
    """

    def __init__(
        self,
        *,
        data_type: DataType = DataType.STRING,
        is_mandatory: bool = True,
        is_unique: bool = False,
        foreign: Optional[set[str]] = None,
        enum_values: Optional[set[str]] = None,
    ):
        self.data_type = data_type
        self.is_mandatory = is_mandatory
        self.is_unique = is_unique
        self.occurrences: set[str] = set()
        self.foreign = foreign
        self.enum_values = enum_values

    def _check_data_type(self, value: str) -> bool:
        """
        Check that the given value can be converted as expected
        """
        convert = _CONVERTERS[self.data_type]
        try:
            convert(value)
        except ValueError:
            return False
        return True

    def _check_enum(self, value: str) -> bool:
        """
        Check that the given value can be converted to ENUM
        Example: gender = {MALE|FEMALE}
        """
        if self.enum_values and value not in self.enum_values:
            return False
        return True

    def _check_unicity(self, value: str) -> bool:
        """
        Check that the given value is unique.
        Add the value to the list of occurrences.
        """
        if not self.is_unique:
            return True
        if value in self.occurrences:
            return False
        self.occurrences.add(value)
        return True

    def _check_foreign(self, value: str) -> bool:
        """
        Check that the given value refers to an existing value
        """
        if not self.foreign:
            return True
        return value in self.foreign

    def check(self, value: Optional[str]) -> Optional[Issue]:
        """
        Do all checks to validate that the given value respect the column rules.
        Return an issue if something's wrong.
        Issues are mutually exclusives => no no need to return a list.
        """
        if value is None or value == "":
            if self.is_mandatory:
                return Issue.MISSING_VALUE
            return None

        if value.strip('"') != value:
            return Issue.REPEATED_QUOTES

        if not self._check_enum(value):
            return Issue.UNAUTHORIZED_VALUE
        if not self._check_data_type(value):
            return Issue.WRONG_DATATYPE
        if not self._check_unicity(value):
            return Issue.DUPLICATE_VALUE
        if not self._check_foreign(value):
            return Issue.UNKNOWN_REFERENCE
        return None



from typing import Optional

from ..abstract import (
    CATALOG_ASSETS,
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
    WarehouseAsset,
)

DB_FILTERED_ASSETS = (
    *CATALOG_ASSETS,
    WarehouseAsset.VIEW_DDL,
)


def _database_filter(db_list: Optional[list[str]], allow: bool) -> str:
    if not db_list:
        return ""
    keyword = "IN" if allow else "NOT IN"
    db_list_str = ", ".join([f"'{db}'" for db in db_list])
    return f"AND database_name {keyword} ({db_list_str})"


def _transient_filter(has_transient: Optional[bool] = False) -> str:
    return "TRUE" if has_transient else "FALSE"


class SnowflakeQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from Snowflake.
    """

    def __init__(
        self,
        time_filter: Optional[TimeFilter] = None,
        db_allowed: Optional[list[str]] = None,
        db_blocked: Optional[list[str]] = None,
        fetch_transient: Optional[bool] = False,
    ):
        super().__init__(time_filter=time_filter)
        self._db_allowed = _database_filter(db_allowed, allow=True)
        self._db_blocked = _database_filter(db_blocked, allow=False)
        self._fetch_transient = _transient_filter(fetch_transient)

    def _apply_db_filters(self, statement: str) -> str:
        statement = statement.format(
            database_allowed=self._db_allowed,
            database_blocked=self._db_blocked,
            has_fetch_transient=self._fetch_transient,
        )

        return statement

    def build(self, asset: WarehouseAsset) -> list[ExtractionQuery]:
        query = self.build_default(asset)

        if asset in DB_FILTERED_ASSETS:
            query.statement = self._apply_db_filters(query.statement)

        return [query]



from typing import Optional


def build_url(host: Optional[str], endpoint: str):
    if not host:
        return endpoint
    if not host.startswith("https://"):
        host = "https://" + host
    return f"{host.strip('/')}/{endpoint}"



import logging
from collections.abc import Iterator
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Any, Optional

import requests

from ....utils import (
    RequestSafeMode,
    at_midnight,
    batch_of_length,
    current_date,
    empty_iterator,
    handle_response,
    past_date,
    retry,
    timestamp_ms,
)
from ..assets import DomoAsset
from .credentials import DomoCredentials
from .endpoints import Endpoint, EndpointFactory
from .pagination import Pagination

RawData = Iterator[dict]

DOMO_PUBLIC_URL = "https://api.domo.com"
DEFAULT_TIMEOUT = 120
TOKEN_EXPIRATION_SECONDS = timedelta(seconds=3000)  # auth token lasts 1 hour


# Safe Mode
VOLUME_IGNORED = 10
IGNORED_ERROR_CODES = (
    HTTPStatus.BAD_REQUEST,
    HTTPStatus.NOT_FOUND,
)
DOMO_SAFE_MODE = RequestSafeMode(
    max_errors=VOLUME_IGNORED,
    status_codes=IGNORED_ERROR_CODES,
)

_RETRY_EXCEPTIONS = [
    requests.exceptions.ConnectTimeout,
    requests.exceptions.ReadTimeout,
]
_RETRY_COUNT = 2
_RETRY_BASE_MS = 10 * 60 * 1000  # 10 minutes

_PARENT_FOLDER = "/Dashboards"

_CARDS_BATCH_SIZE = 100

logger = logging.getLogger(__name__)


class DomoClient:
    """
    Connect to Domo API and fetch main assets.
    https://developer.domo.com/portal/8ba9aedad3679-ap-is#platform-oauth-apis
    """

    def __init__(
        self,
        credentials: DomoCredentials,
        safe_mode: Optional[RequestSafeMode] = None,
    ):
        self._authentication = credentials.authentication
        self._bearer_headers: Optional[dict] = None
        self._session = requests.session()
        self._token_creation_time: datetime = datetime.min
        self._endpoint_factory = EndpointFactory(credentials.base_url)
        self._private_headers = credentials.private_headers
        self._timeout = DEFAULT_TIMEOUT
        self.base_url = credentials.base_url
        self.cloud_id = credentials.cloud_id
        self.safe_mode = safe_mode or DOMO_SAFE_MODE

    def _token_expired(self) -> bool:
        token_lifetime = datetime.now() - self._token_creation_time
        return token_lifetime > TOKEN_EXPIRATION_SECONDS

    def _bearer_auth(self) -> dict:
        if self._bearer_headers and not self._token_expired():
            return self._bearer_headers

        logger.info("Refreshing authentication token...")

        basic_authentication = self._authentication
        endpoint = self._endpoint_factory.authentication

        response = self._session.get(
            endpoint.url(),
            auth=basic_authentication,
            timeout=self._timeout,
        )
        response.raise_for_status()
        result = response.json()

        bearer_token = result["access_token"]
        self._bearer_headers = {"authorization": f"Bearer {bearer_token}"}
        self._token_creation_time = datetime.now()

        return self._bearer_headers

    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_COUNT,
        base_ms=_RETRY_BASE_MS,
    )
    def _get(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> Any:
        params = params if params else {}
        is_private = endpoint.is_private
        headers = self._private_headers if is_private else self._bearer_auth()

        response = self._session.get(
            url=endpoint.url(asset_id),
            headers=headers,
            params=params,
            timeout=self._timeout,
        )

        return handle_response(response, self.safe_mode)

    def _get_element(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> dict:
        """Used when the response only contains one element"""
        return self._get(endpoint, params, asset_id)

    def _get_many(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> list[dict]:
        """Used when the response contains multiple elements"""
        return self._get(endpoint, params, asset_id)

    def _get_paginated(self, endpoint: Endpoint) -> list[dict]:
        """Used when the response is paginated and need iterations"""
        pagination = Pagination()
        all_results: list[dict] = []

        while pagination.needs_increment:
            params = {**pagination.params, **endpoint.params}
            results = self._get_many(endpoint=endpoint, params=params)
            all_results.extend(results)
            number_of_items = len(results)
            pagination.increment_offset(number_of_items)

        return all_results

    def _cards_metadata(self, card_ids: list[int]) -> Iterator[dict]:
        # batch to avoid hitting the URL max length
        for batch_card_ids in batch_of_length(card_ids, _CARDS_BATCH_SIZE):
            endpoint = self._endpoint_factory.cards_metadata(batch_card_ids)
            yield from self._get_element(endpoint)

    def _datasources(self, card_ids: list[int]) -> RawData:
        """Yields all distinct datasources associated to the given cards"""
        if not card_ids:
            return empty_iterator()

        processed: set[str] = set()
        for card in self._cards_metadata(card_ids):
            for datasource in card["datasources"]:
                id_ = datasource["dataSourceId"]
                if id_ in processed:
                    continue
                yield {
                    "id": id_,
                    "name": datasource["dataSourceName"],
                    "type": datasource["dataType"],
                }
                processed.add(id_)

    def _process_pages(
        self,
        page_tree: list[dict],
        parent_path: str = _PARENT_FOLDER,
    ) -> Iterator[dict]:
        """Recursively fetch pages while building the folder architecture"""
        if not page_tree:
            return empty_iterator()

        for page in page_tree:
            page_id = page.get("id")
            page_name = page.get("name")
            if not (page_id and page_name):
                logger.debug(f"This page has no id nor name: {page}")
                continue

            page_children = page.get("children", list)

            detail = self._get_element(
                self._endpoint_factory.pages,
                asset_id=page_id,
            )

            if not detail:
                continue

            datasources = self._datasources(detail.get("cardIds", []))
            yield {
                **detail,
                "datasources": list(datasources),
                "path": parent_path,
                "base_url": self.base_url,
            }
            yield from self._process_pages(
                page_tree=page_children,
                parent_path=f"{parent_path}/{page_name}",
            )

    def _link_table_to_dataset(self, dataset: dict, dataset_id: str) -> dict:
        if not self.cloud_id:
            return dataset

        lineage_endpoint = self._endpoint_factory.table_lineage(
            dataset_id=dataset_id,
            cloud_id=self.cloud_id,
        )
        dataset_table_lineage = self._get_element(lineage_endpoint)

        return {**dataset, **dataset_table_lineage}

    def _pages(self) -> RawData:
        page_hierarchy = self._get_paginated(self._endpoint_factory.pages)
        yield from self._process_pages(page_hierarchy)

    def _datasets(self) -> RawData:
        dataset_list = self._get_paginated(self._endpoint_factory.datasets)

        for dataset in dataset_list:
            dataset_id = dataset.get("id")
            if not dataset_id:
                continue

            dataset_detail = self._get_element(
                endpoint=self._endpoint_factory.datasets,
                asset_id=dataset_id,
            )

            yield {
                **self._link_table_to_dataset(
                    dataset=dataset_detail,
                    dataset_id=dataset_id,
                ),
                "base_url": self.base_url,
            }

    def _unique_datasets(self) -> RawData:
        ids_encountered: set[str] = set()
        for dataset in self._datasets():
            dataset_id = dataset.get("id")
            if not dataset_id or dataset_id in ids_encountered:
                continue
            ids_encountered.add(dataset_id)
            yield dataset

    def _users(self) -> RawData:
        user_list = self._get_paginated(self._endpoint_factory.users)
        for user in user_list:
            user_id = user.get("id")
            if not user_id:
                continue
            yield self._get_element(
                endpoint=self._endpoint_factory.users,
                asset_id=user_id,
            )

    def _audit(self) -> RawData:
        yesterday = timestamp_ms(at_midnight(past_date(1)))
        today = timestamp_ms(at_midnight(current_date()))
        yield from self._get_paginated(
            self._endpoint_factory.audit(yesterday, today)
        )

    def _dataflows(self) -> RawData:
        dataflows = self._get_many(self._endpoint_factory.dataflows)
        for dataflow in dataflows:
            dataflow_id = dataflow.get("id")
            if not dataflow_id:
                continue
            yield self._get_element(self._endpoint_factory.lineage(dataflow_id))

    def fetch(self, asset: DomoAsset) -> RawData:
        """Returns the needed metadata for the queried asset"""
        if asset == DomoAsset.PAGES:
            yield from self._pages()

        elif asset == DomoAsset.DATASETS:
            yield from self._unique_datasets()

        elif asset == DomoAsset.USERS:
            yield from self._users()

        elif asset == DomoAsset.AUDIT:
            yield from self._audit()

        elif asset == DomoAsset.DATAFLOWS:
            yield from self._dataflows()

        else:
            raise ValueError(f"This asset {asset} is unknown")



import pytest

from .client_metadata_api import _TIMEOUT_MESSAGE, _check_errors, _deduplicate
from .errors import TableauApiError, TableauApiTimeout


def test__deduplicate():
    result_pages = iter(
        [
            [
                {"id": 1, "name": "workbook_1"},
                {"id": 2, "name": "workbook_2"},
            ],
            [
                {"id": 1, "name": "workbook_1"},
                {"id": 3, "name": "workbook_3"},
                {"id": 4, "name": "workbook_4"},
            ],
            [
                {"id": 4, "name": "workbook_4"},
                {"id": 5, "name": "workbook_5"},
                {"id": 5, "name": "workbook_5"},
                {"id": 5, "name": "workbook_5"},
            ],
            [
                {"id": 1, "name": "workbook_1"},
                {"id": 3, "name": "workbook_3"},
            ],
        ]
    )
    deduplicated = _deduplicate(result_pages)
    assert len(deduplicated) == 5
    deduplicated_keys = {item["id"] for item in deduplicated}
    assert deduplicated_keys == {1, 2, 3, 4, 5}


def _timeout_error():
    return {"message": _TIMEOUT_MESSAGE}


def _critical_error():
    return {
        "message": "critical error",
        "extensions": {"severity": "Error"},
    }


def _unknown_severity_error():
    return {
        "message": "unknown severity error",
        "extensions": {"severity": None},
    }


def _warning():
    return {
        "message": "critical error",
        "extensions": {"severity": "Warning"},
    }


def test__check_errors():
    answer = {"data": []}  # no errors
    assert _check_errors(answer) is None

    # only warnings
    answer = {
        "data": [],
        "errors": [
            _warning(),
            _warning(),
            _warning(),
        ],
    }
    assert _check_errors(answer) is None

    # timeout issues should be prio (because they can be retried)
    answer = {
        "data": [],
        "errors": [
            _warning(),
            _critical_error(),
            _timeout_error(),
            _warning(),
        ],
    }
    with pytest.raises(TableauApiTimeout):
        _check_errors(answer)

    # expect critical error
    answer = {
        "data": [],
        "errors": [
            _warning(),
            _warning(),
            _critical_error(),
        ],
    }
    with pytest.raises(TableauApiError):
        _check_errors(answer)

    # unknown severity is considered as critical
    answer = {
        "data": [],
        "errors": [
            _unknown_severity_error(),
            _warning(),
        ],
    }
    with pytest.raises(TableauApiError):
        _check_errors(answer)



from collections.abc import Iterator
from functools import partial
from typing import Callable

from ....utils import (
    APIClient,
    BasicAuth,
    fetch_all_pages,
)
from ..assets import SodaAsset
from .credentials import SodaCredentials
from .endpoints import SodaEndpointFactory
from .pagination import SodaCloudPagination

_REQUESTS_PER_MINUTE = 10
_SECONDS_PER_MINUTE = 60
_RATE_LIMIT_MS = (_SECONDS_PER_MINUTE // _REQUESTS_PER_MINUTE) + 1
_CLOUD_PAGE_SIZE = 100

HEADERS = {"Content-Type": "application/json"}


class SodaClient(APIClient):
    def __init__(self, credentials: SodaCredentials):
        cloud_auth = BasicAuth(
            username=credentials.api_key, password=credentials.secret
        )
        super().__init__(
            host=credentials.host, auth=cloud_auth, headers=HEADERS
        )

    @property
    def _base_request(self) -> Callable:
        return partial(self._get, params={"size": _CLOUD_PAGE_SIZE})

    def datasets(self) -> Iterator[dict]:
        request = partial(
            self._base_request, endpoint=SodaEndpointFactory.datasets()
        )
        yield from fetch_all_pages(
            request, SodaCloudPagination, rate_limit=_RATE_LIMIT_MS
        )

    def checks(self) -> Iterator[dict]:
        request = partial(
            self._base_request, endpoint=SodaEndpointFactory.checks()
        )
        yield from fetch_all_pages(
            request, SodaCloudPagination, rate_limit=_RATE_LIMIT_MS
        )

    def fetch(self, asset: SodaAsset) -> Iterator:
        if asset == SodaAsset.DATASETS:
            return self.datasets()
        if asset == SodaAsset.CHECKS:
            return self.checks()
        raise ValueError(f"The asset {asset}, is not supported")



#!/usr/bin/env python3
import logging
import ntpath
from collections.abc import Iterable
from datetime import datetime
from typing import Optional
from uuid import UUID

import requests

from ..utils.retry import retry
from .constant import (
    INGEST_URLS,
    PATH_TEMPLATES,
    RETRY_BASE_MS,
    RETRY_JITTER_MS,
    RETRY_STRATEGY,
    FileType,
)
from .enums import Zone
from .env import get_blob_env
from .settings import UploaderSettings
from .utils import iter_files

logger = logging.getLogger(__name__)

_EXCEPTIONS = (
    requests.exceptions.Timeout,
    requests.exceptions.ConnectTimeout,
)


def _path_and_url(
    source_id: UUID,
    file_type: FileType,
    file_path: str,
    zone: Zone,
) -> tuple[str, str]:
    now = datetime.utcnow()
    timestamp = int(now.timestamp())
    filename = ntpath.basename(file_path)
    path_template = PATH_TEMPLATES[file_type]
    path = path_template.format(
        timestamp=timestamp,
        source_id=source_id,
        filename=filename,
    )

    url = f"{INGEST_URLS[zone]}/{path}"

    return path, url


def _headers(token: str) -> dict:
    return {
        "Authorization": f"Token {token}",
        "Accept": "text/csv, application/json",
    }


def _upload(
    token: str,
    source_id: UUID,
    file_path: str,
    file_type: FileType,
    zone: Optional[Zone] = Zone.EU,
) -> None:
    """
    Upload the given file to Google Cloud Storage (GCS)
    - Don't call GCS API directly
    - Call the ingestion proxy which handles authorisation and uploading
    """
    if not zone:
        zone = Zone.EU
    path, url = _path_and_url(source_id, file_type, file_path, zone)
    headers = _headers(token)
    timeout, max_retries = get_blob_env()

    with open(file_path, "rb") as file_content:

        @retry(
            exceptions=_EXCEPTIONS,
            max_retries=max_retries,
            base_ms=RETRY_BASE_MS,
            jitter_ms=RETRY_JITTER_MS,
            strategy=RETRY_STRATEGY,
        )
        def _request_post():
            response = requests.post(
                url=url,
                headers=headers,
                files={"file": file_content},
                timeout=timeout,
            )
            response.raise_for_status()

        _request_post()

    logger.info(f"Uploaded {file_path} as {file_type.value} to {path}")


def upload_manifest(
    token: str,
    source_id: UUID,
    zone: Optional[Zone],
    file_path: Optional[str] = None,
) -> None:
    """
    token: backend public API token
    source_id: id for the source
    file_path: path to the local manifest to upload
    """
    if not file_path:
        raise ValueError("file path is needed to upload a manifest")
    _upload(
        token=token,
        source_id=source_id,
        file_path=file_path,
        file_type=FileType.DBT,
        zone=zone,
    )


def upload(
    token: str,
    source_id: UUID,
    file_type: FileType,
    zone: Optional[Zone],
    file_path: Optional[str] = None,
    directory_path: Optional[str] = None,
) -> None:
    """
    token: backend public API token
    source_id: id for the source
    file_type: type of file(s) uploaded - see FileType Enum
    file_path: path to the local visualization or warehouse file to upload
    directory_path: path to the local directory containing files to upload
    """
    files: Iterable[str]
    if directory_path:
        files = iter_files(directory_path)
    elif file_path:
        files = [file_path]
    else:
        message = "either file_path or directory_path should be defined"
        raise ValueError(message)

    for file_ in files:
        _upload(
            token=token,
            source_id=source_id,
            file_path=file_,
            file_type=file_type,
            zone=zone,
        )


def upload_any(**kwargs) -> None:
    """
    entrypath to upload either a file or a manifest

    token: backend public API token
    source_id: id for the source
    file_type: type of file(s) uploaded - see FileType Enum
    file_path: path to the local visualization or warehouse file to upload
    directory_path: path to the local directory containing files to upload
    """

    settings = UploaderSettings(**kwargs)
    file_type = settings.file_type

    if file_type == FileType.DBT:
        assert not settings.directory_path
        upload_manifest(
            token=settings.token,
            source_id=settings.source_id,
            file_path=settings.file_path,
            zone=settings.zone,
        )
        return None

    upload(
        token=settings.token,
        source_id=settings.source_id,
        file_type=file_type,
        file_path=settings.file_path,
        directory_path=settings.directory_path,
        zone=settings.zone,
    )



import os
from typing import Optional

from .constant import (
    DEFAULT_RETRY,
    DEFAULT_TIMEOUT,
    ENVIRON_RETRY,
    ENVIRON_TIMEOUT,
)


def _parse_float(value: Optional[str], default: float) -> float:
    if not value:
        return default
    try:
        return float(value)
    except ValueError:
        return default


def _parse_int(value: Optional[str], default: int) -> int:
    if not value:
        return default
    try:
        return int(value)
    except ValueError:
        return default


def get_blob_env() -> tuple[float, int]:
    """
    Retrieve timeout and retries values. It look for environment variables
    first and return default value otherwise
    """
    timeout = _parse_float(os.environ.get(ENVIRON_TIMEOUT), DEFAULT_TIMEOUT)
    retry = _parse_int(os.environ.get(ENVIRON_RETRY), DEFAULT_RETRY)

    return timeout, retry



from typing import Optional

from pydantic import AliasChoices, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

METABASE_DB_ENV_PREFIX = "CASTOR_METABASE_DB_"


class MetabaseDbCredentials(BaseSettings):
    """Metabase's credentials to connect to Metabase DB"""

    model_config = SettingsConfigDict(
        env_prefix=METABASE_DB_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    host: str
    port: str
    database: str
    schema_: str = Field(
        validation_alias=AliasChoices(
            f"{METABASE_DB_ENV_PREFIX}SCHEMA", "schema"
        )
    )
    user: str = Field(validation_alias=f"{METABASE_DB_ENV_PREFIX}USERNAME")
    password: str = Field(repr=False)

    encryption_secret_key: Optional[str] = Field(repr=False, default=None)
    require_ssl: Optional[bool] = Field(
        validation_alias=f"{METABASE_DB_ENV_PREFIX}REQUIRE_SSL_KEY"
    )



import logging
from abc import ABC, abstractmethod
from typing import Optional, Union

from requests.auth import AuthBase, HTTPBasicAuth

logger = logging.getLogger(__name__)


class BasicAuth(HTTPBasicAuth):
    """
    Authentication for API using basic auth method

    - Instantiate with username and password
    - pass the Auth class to the APIClient
    """

    def refresh_token(self):
        pass


class CustomAuth(AuthBase, ABC):
    """
    Authentication for API using custom auth method

    You need to:
     - implement the `_authentication_header()` method
     - pass the Auth class to the APIClient
    """

    def refresh_token(self):
        """Method to refresh the token if token expires"""
        pass

    @abstractmethod
    def _authentication_header(self) -> dict[str, str]:
        pass

    def __call__(self, r):
        r.headers = {**r.headers, **self._authentication_header()}
        return r


class BearerAuth:
    """
    Authentication for API using Bearer tokens

    You need to:
    - implement the `fetch_token()` method
    - pass the Auth class to the APIClient
    """

    _token: Optional[str] = None
    authentication_key = "Authorization"

    @abstractmethod
    def fetch_token(self) -> Optional[str]:
        """Method that should return the bearer token"""
        pass

    def refresh_token(self):
        """Method to refresh the token if token expires"""
        self._fetch_token(force_refresh=True)

    def _fetch_token(self, force_refresh: bool = False) -> Optional[str]:
        if not self._token or force_refresh:
            logger.info("Refreshing authentication token...")
            self._token = self.fetch_token()
        return f"Bearer {self._token}"

    def __call__(self, r):
        r.headers[self.authentication_key] = self._fetch_token()
        return r


Auth = Union[BasicAuth, CustomAuth, BearerAuth]



from datetime import date, datetime

from .time import at_midnight, date_after, format_rfc_3339_date, timestamp_ms


def test_at_midnight():
    assert at_midnight(date(1989, 2, 2)) == datetime(1989, 2, 2, 0, 0, 0)


def test_date_after():
    day = date(1999, 12, 31)
    assert date_after(day, 3) == date(2000, 1, 3)


def test_timestamp_ms():
    dt = datetime(1991, 4, 3, 0, 0)
    result = timestamp_ms(dt)
    expected = 670636800000
    assert result == expected


def test_format_rfc_3339_date():
    dt = datetime(1995, 4, 3, 2, 1)
    result = format_rfc_3339_date(dt)
    expected = "1995-04-03T02:01:00Z"
    assert result == expected



"""convert files to csv"""

import csv
import ctypes
import json
import logging
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator, Sequence
from datetime import date, datetime
from enum import Enum
from typing import IO, Any, Union
from uuid import UUID

from ..types import CsvOptions

CSV_OPTIONS: CsvOptions = {
    "delimiter": ",",
    # bigquery import requires explicit quoting
    "quoting": csv.QUOTE_ALL,
    "quotechar": '"',
}

CSV_FIELD_SIZE_MB = 100

ScalarValue = Union[int, float, None, str]

logger = logging.getLogger(__name__)


def _header(row: dict) -> Sequence[str]:
    return [str(r) for r in row.keys()]


def _scalar(value: Any) -> ScalarValue:
    if isinstance(value, str):
        if "\x00" in value:  # infrequent error caused by bad encoding
            value = remove_unsupported_byte(value)
            logger.warning("Removed unsupported byte to write to csv")
            return value

        return value
    if isinstance(value, (int, float)):
        return value
    if isinstance(value, (date, datetime)):
        return value.isoformat()
    if value is None:
        return None
    if isinstance(value, (dict, list)):
        return json.dumps(value)
    # fallback
    return str(value)


def _row(header: Sequence[str], row: dict) -> list[ScalarValue]:
    return [_scalar(row.get(h)) for h in header]


def remove_unsupported_byte(element: ScalarValue) -> ScalarValue:
    if not isinstance(element, str):
        return element

    return re.sub("\x00", "", element)


def to_string_array(arr_json: str) -> list[str]:
    """
    Converts a JSON-serialized string array value as a string to a list
    Ex: '["items","count"]' to ["items", "order"]
    """

    if not arr_json.startswith("[") or not arr_json.endswith("]"):
        raise ValueError(f"Cannot deserialize (not an array): {arr_json}")

    try:
        array = json.loads(arr_json)
    except ValueError:
        raise ValueError(f"Cannot deserialize (not a JSON): {arr_json}")

    if not all(isinstance(el, str) for el in array):
        raise ValueError(f"Not an array of strings: {arr_json}")

    return array


def to_csv(buffer: IO[str], data: Iterable[dict]) -> bool:
    """convert data as list of dicts to CSV string"""

    writer = csv.writer(buffer, **CSV_OPTIONS)

    header = None
    for row in data:
        # write header once
        if not header:
            header = _header(row)
            writer.writerow(header)
        converted = _row(header, row)
        writer.writerow(converted)
    return True


def from_csv(buffer: IO[str]) -> Iterator[dict]:
    """convert data as from a CSV string to list of dict"""
    try:
        reader = csv.reader(buffer, **CSV_OPTIONS)
        header: list[str] = []
        for row in reader:
            if not header:
                header = list(row)
                continue
            yield {h: v for h, v in zip(header, row)}
    finally:
        # closing of the file must happen after all iterations
        buffer.close()


class Formatter(ABC):
    """
    Abstract class to Serialize/Deserialize to any format
    """

    @abstractmethod
    def extension(self) -> str:
        pass

    @staticmethod
    @abstractmethod
    def serialize(buffer: IO[str], data: Iterable[dict]) -> bool:
        pass

    @staticmethod
    @abstractmethod
    def deserialize(data: IO[str]) -> Iterator[dict]:
        pass


def _set_csv_field_size_limit(target_limit_mb: int) -> None:
    """
    Safely set the maximum CSV field size limit across platforms.

    This function wraps `csv.field_size_limit()` to avoid `OverflowError` on
    Windows, where the maximum C long is only 32 bits (2^31 - 1). On Linux and
    macOS, the C long is typically 64 bits, allowing much larger values.

    The requested limit is specified in megabytes and converted to bytes.

    It is then clamped to the maximum value supported by:
      - the current platform's C long
      - Python's `sys.maxsize`
      - the requested target limit
    """
    target_limit_bytes = target_limit_mb * 1024**2

    # max value of C long for the current platform
    platform_c_long = (1 << (8 * ctypes.sizeof(ctypes.c_long) - 1)) - 1

    limit_bytes = min(target_limit_bytes, sys.maxsize, platform_c_long)

    csv.field_size_limit(limit_bytes)


class CsvFormatter(Formatter):
    """
    Serialize/Deserialize CSV
    """

    def extension(self) -> str:
        return "csv"

    # increase the size limit (some fields are very large)
    _set_csv_field_size_limit(target_limit_mb=CSV_FIELD_SIZE_MB)

    @staticmethod
    def serialize(buffer: IO[str], data: Iterable[dict]) -> bool:
        return to_csv(buffer, data)

    @staticmethod
    def deserialize(buffer: IO[str]) -> Iterator[dict]:
        return from_csv(buffer)


class CustomEncoder(json.JSONEncoder):
    """supersedes the default encoder to handle additional types"""

    def default(self, obj: object) -> object:
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, Enum):
            return obj.name
        return json.JSONEncoder.default(self, obj)


class JsonFormatter(Formatter):
    """
    Serialize/Deserialize JSON
    """

    def extension(self) -> str:
        return "json"

    @staticmethod
    def serialize(buffer: IO[str], data: Iterable[dict]) -> bool:
        try:
            json.dump(data, fp=buffer, cls=CustomEncoder)
            return True
        except ValueError:
            return False

    @staticmethod
    def deserialize(buffer: IO[str]) -> Iterator[dict]:
        try:
            data = json.load(buffer)
            return data
        finally:
            buffer.close()



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_NOTION_"


class NotionCredentials(BaseSettings):
    """Class to handle Notion rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    token: str = Field(repr=False)



from ...types import ExternalAsset


class DomoAsset(ExternalAsset):
    """Domo assets"""

    AUDIT = "audit"
    DATAFLOWS = "dataflows"
    DATASETS = "datasets"
    PAGES = "pages"
    USERS = "users"



import logging
import sys
import warnings


def deprecate_python(min_version_supported: tuple[int, ...]):
    """raises a warning if python version < min_version_supported"""

    python_version = (
        sys.version_info.major,
        sys.version_info.minor,
        sys.version_info.micro,
    )

    python_version_str = ".".join(map(str, python_version))
    min_supported_str = ".".join(map(str, min_version_supported))

    message = f"You are using python version {python_version_str}, please upgrade to version {min_supported_str} or higher."
    " Your version will be soon deprecated"

    if python_version < min_version_supported:
        warnings.warn(message, DeprecationWarning)

        # Since warnings are disabled by default, let's add a log as well
        logging.warning(message)



from datetime import date, datetime, timedelta, timezone
from typing import Union

ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"


def current_datetime() -> datetime:
    """Returns the current datetime"""
    return datetime.utcnow()


def current_date() -> date:
    """Returns the current datetime"""
    return current_datetime().date()


def current_timestamp() -> int:
    """
    Returns the current timestamp from epoch (rounded to the nearest second)
    """
    return int(current_datetime().timestamp())


def _set_uct_timezone(ts: datetime) -> datetime:
    if ts.tzinfo is None:
        return ts.replace(tzinfo=timezone.utc)
    return ts


def timestamp_ms(ts: datetime) -> int:
    """
    Return ts timestamp in millisecond (rounded)
    """
    ts_utc = _set_uct_timezone(ts)
    return int(ts_utc.timestamp() * 1000)


def now(tz: bool = False) -> datetime:
    """
    provide current time
    optionally localize timezone
    """
    ts = datetime.utcnow()
    return _set_uct_timezone(ts) if tz else ts


def past_date(past_days: int) -> date:
    """returns a date in the past"""
    return now().date() - timedelta(past_days)


def at_midnight(date_: date) -> datetime:
    """convert date into datetime at midnight: 00:00:00"""
    return datetime.combine(date_, datetime.min.time())


def date_after(day: date, future_days: int) -> date:
    """returns the date `future_days` after `day`"""
    return day + timedelta(future_days)


def format_date(timestamp: Union[datetime, date]) -> str:
    return timestamp.strftime(ISO_FORMAT)


def format_rfc_3339_date(timestamp: datetime) -> str:
    return timestamp.isoformat(timespec="seconds") + "Z"


def yesterday() -> date:
    return current_date() - timedelta(days=1)



import logging
from functools import partial
from typing import Optional

from tqdm import tqdm  # type: ignore

from ...utils import fetch_all_pages
from ...utils.salesforce import SalesforceBaseClient, SalesforceCredentials
from .format import SalesforceFormatter
from .pagination import SalesforceSQLPagination, format_sobject_query
from .soql import (
    DESCRIPTION_QUERY_TPL,
    SOBJECT_FIELDS_QUERY_TPL,
)

logger = logging.getLogger(__name__)


class SalesforceClient(SalesforceBaseClient):
    """
    Salesforce API client to extract sobjects
    """

    def __init__(self, credentials: SalesforceCredentials):
        super().__init__(credentials)
        self.formatter = SalesforceFormatter()

    @staticmethod
    def name() -> str:
        return "Salesforce"

    def fetch_sobjects(self) -> list[dict]:
        """Fetch all sobjects"""
        logger.info("Extracting sobjects")
        query = format_sobject_query()
        request_ = partial(
            self._get, endpoint=self.query_endpoint, params={"q": query}
        )
        results = fetch_all_pages(request_, SalesforceSQLPagination)
        return list(results)

    def fetch_fields(self, sobject_name: str) -> list[dict]:
        """Fetches fields of a given sobject"""
        query = SOBJECT_FIELDS_QUERY_TPL.format(
            entity_definition_id=sobject_name
        )
        response = self._get(self.tooling_endpoint, params={"q": query})
        return response["records"]

    def fetch_description(self, table_name: str) -> Optional[str]:
        """Retrieve description of a table"""
        query = DESCRIPTION_QUERY_TPL.format(table_name=table_name)
        response = self._get(self.tooling_endpoint, params={"q": query})
        if not response["records"]:
            return None
        return response["records"][0]["Description"]

    def add_table_descriptions(self, sobjects: list[dict]) -> list[dict]:
        """
        Add table descriptions.
        We use the tooling API which does not handle well the LIMIT in SOQL
        so we have to retrieve descriptions individually
        """
        described_sobjects = []
        for sobject in sobjects:
            description = self.fetch_description(sobject["QualifiedApiName"])
            described_sobjects.append({**sobject, "Description": description})
        return described_sobjects

    def tables(self) -> list[dict]:
        """
        Get Salesforce sobjects as tables
        """
        sobjects = self.fetch_sobjects()
        logger.info(f"Extracted {len(sobjects)} sobjects")
        described_sobjects = self.add_table_descriptions(sobjects)
        return list(self.formatter.tables(described_sobjects))

    def columns(
        self, sobject_names: list[tuple[str, str]], show_progress: bool = True
    ) -> list[dict]:
        """
        Get salesforce sobject fields as columns
        show_progress: optionally deactivate the tqdm progress bar
        """
        sobject_fields: dict[str, list[dict]] = dict()
        for api_name, table_name in tqdm(
            sobject_names, disable=not show_progress
        ):
            fields = self.fetch_fields(api_name)
            sobject_fields[table_name] = fields
        return list(self.formatter.columns(sobject_fields))



from unittest.mock import Mock

from .sources_transformer import SigmaSourcesTransformer

_ALL_SOURCES = [
    {
        "asset_id": "asset1",
        "sources": [
            {"type": "dataset", "inodeId": "1234"},  # non-table source
            {"type": "table", "inodeId": "table1"},
            {"type": "table", "inodeId": "table2"},
        ],
    },
    {
        "asset_id": "asset2",
        "sources": [
            {"type": "table", "inodeId": "table1"},  # repeated source
        ],
    },
]


_TABLE_TO_PATH = {
    "table1": {
        "connectionId": "conn1",
        "path": ["db", "schema", "table1"],
    },
    "table2": {
        "connectionId": "conn2",
        "path": ["db", "schema", "table2"],
    },
}


def test__map_table_id_to_connection_path():
    transformer = SigmaSourcesTransformer(api_client=Mock())

    def mock_get(endpoint):
        if "table1" in endpoint:
            return _TABLE_TO_PATH["table1"]
        elif "table2" in endpoint:
            return _TABLE_TO_PATH["table2"]
        else:
            raise ValueError(f"Unexpected endpoint: {endpoint}")

    transformer.api_client._get.side_effect = mock_get

    result = transformer._map_table_id_to_connection_path(_ALL_SOURCES)

    assert len(result) == 2
    assert result["table1"] == {
        "connectionId": "conn1",
        "path": ["db", "schema", "table1"],
    }
    assert result["table2"] == {
        "connectionId": "conn2",
        "path": ["db", "schema", "table2"],
    }
    assert transformer.api_client._get.call_count == 2


def test__transform_sources():
    transformer = SigmaSourcesTransformer(api_client=Mock())

    result = list(transformer._transform_sources(_ALL_SOURCES, _TABLE_TO_PATH))

    assert len(result) == 2

    asset_1_results = result[0]
    assert len(asset_1_results["sources"]) == 3
    actual_sources = sorted(
        asset_1_results["sources"], key=lambda x: x["inodeId"]
    )
    expected_sources = [
        {"type": "dataset", "inodeId": "1234"},
        {
            "type": "table",
            "inodeId": "table1",
            "connectionId": "conn1",
            "path": ["db", "schema", "table1"],
        },
        {
            "type": "table",
            "inodeId": "table2",
            "connectionId": "conn2",
            "path": ["db", "schema", "table2"],
        },
    ]
    expected_sources = sorted(expected_sources, key=lambda x: x["inodeId"])
    assert actual_sources == expected_sources

    asset_2_results = result[1]
    assert asset_2_results["asset_id"] == "asset2"
    assert asset_2_results["sources"] == [
        {
            "type": "table",
            "inodeId": "table1",
            "connectionId": "conn1",
            "path": ["db", "schema", "table1"],
        }
    ]



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_DBT_CLOUD_URL = "https://cloud.getdbt.com"


class DbtCredentials(BaseSettings):
    """dbt credentials: host has default value"""

    host: str = Field(
        default=DEFAULT_DBT_CLOUD_URL, validation_alias="CASTOR_DBT_HOST"
    )
    job_id: str = Field(..., validation_alias="CASTOR_DBT_JOB_ID")
    token: str = Field(..., validation_alias="CASTOR_DBT_TOKEN")
    model_config = SettingsConfigDict(extra="ignore", populate_by_name=True)



from .credentials import DEFAULT_SCOPE, PowerbiCredentials


def test_credentials():
    tenant_id = "any_tenant_id"
    client_id = "any_client_id"
    secret = "🤫"

    # no scopes provided
    credentials = PowerbiCredentials(
        tenant_id=tenant_id,
        client_id=client_id,
        secret=secret,
    )
    assert credentials.scopes == [DEFAULT_SCOPE]

    credentials = PowerbiCredentials(
        tenant_id=tenant_id,
        client_id=client_id,
        secret=secret,
        scopes=None,
    )
    assert credentials.scopes == [DEFAULT_SCOPE]

    # empty scopes
    credentials = PowerbiCredentials(
        tenant_id=tenant_id,
        client_id=client_id,
        secret=secret,
        scopes=[],
    )
    assert credentials.scopes == []

    # with scopes
    scopes = ["foo"]
    credentials = PowerbiCredentials(
        tenant_id=tenant_id,
        client_id=client_id,
        secret=secret,
        scopes=scopes,
    )
    assert credentials.scopes == scopes



from ...utils import SqlalchemyClient, uri_encode

SERVER_URI = "{user}:{password}@{host}:{port}"
MYSQL_URI = f"mysql+pymysql://{SERVER_URI}"
DEFAULT_PORT = 3306

_REQUIRED_KEYS = ("user", "password", "host")


def _check_required_keys(credentials: dict) -> None:
    for key in _REQUIRED_KEYS:
        if key not in credentials:
            raise KeyError(f"Missing {key} in credentials")


class MySQLClient(SqlalchemyClient):
    """MySQL client"""

    @staticmethod
    def name() -> str:
        return "MySQL"

    def _engine_options(self, credentials: dict) -> dict:
        return {}

    def _build_uri(self, credentials: dict) -> str:
        _check_required_keys(credentials)
        uri = MYSQL_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            host=credentials["host"],
            port=credentials.get("port") or DEFAULT_PORT,
        )
        return uri



from argparse import ArgumentParser


def parse_filled_arguments(parser: ArgumentParser) -> dict:
    """Parse arguments and remove all those with None values"""
    parsed_arguments = vars(parser.parse_args())
    return {k: v for k, v in parsed_arguments.items() if v is not None}



import logging
from collections.abc import Iterator
from typing import Any, Callable, Optional
from urllib.parse import urlparse

from mstrio.connection import Connection  # type: ignore
from mstrio.helpers import IServerError  # type: ignore
from mstrio.modeling import (  # type: ignore
    Attribute,
    LogicalTable,
    PhysicalTable,
    PhysicalTableType,
    list_attributes,
    list_facts,
    list_logical_tables,
    list_metrics,
)
from mstrio.project_objects import (  # type: ignore
    Report,
    list_dashboards,
    list_documents,
    list_olap_cubes,
    list_reports,
)
from mstrio.server import Environment  # type: ignore
from mstrio.users_and_groups import User, list_users  # type: ignore
from mstrio.utils.entity import Entity  # type: ignore

from ..assets import StrategyAsset
from .credentials import StrategyCredentials
from .properties import (
    column_properties,
    format_url,
    list_dependencies,
    lookup_table_id,
    safe_get_property,
)

logger = logging.getLogger(__name__)


class StrategyClient:
    """Connect to Strategy through mstrio-py and fetch main assets."""

    def __init__(self, credentials: StrategyCredentials):
        self.base_url = credentials.base_url
        self.connection = Connection(
            base_url=self.base_url,
            username=credentials.username,
            password=credentials.password,
            verbose=False,
        )

        self.hostname = urlparse(self.base_url).hostname

        if credentials.project_ids:
            self.project_ids = credentials.project_ids
        else:
            env = Environment(connection=self.connection)
            self.project_ids = [project.id for project in env.list_projects()]

    def close(self):
        self.connection.close()

    def _common_entity_properties(
        self,
        entity: Entity,
        with_url: bool = True,
        with_description: bool = True,
    ) -> dict:
        """
        Returns the entity's properties, including its dependencies
        and optional URL and/or description.
        """
        dependencies = list_dependencies(entity)
        owner_id = entity.owner.id if isinstance(entity.owner, User) else None
        properties = {
            "dependencies": dependencies,
            "id": entity.id,
            "location": entity.location,
            "name": entity.name,
            "owner_id": owner_id,
            "subtype": entity.subtype,
            "type": entity.type.value,
        }

        if with_url:
            assert self.hostname
            properties["url"] = format_url(
                entity=entity, hostname=self.hostname
            )

        if with_description:
            properties["description"] = safe_get_property(entity, "description")

        return properties

    def _attributes_properties(self, attribute: Attribute) -> dict[str, Any]:
        """
        Attributes have a lookup table, which we need to compute the table lineage.
        """
        return {
            **self._common_entity_properties(attribute, with_url=False),
            "lookup_table_id": lookup_table_id(attribute),
        }

    def _physical_table_properties(
        self, table: Optional[PhysicalTable]
    ) -> Optional[dict[str, Any]]:
        """
        Returns the properties of the physical table, including its columns.
        A physical table can have 1 of these types:
          * "normal": meaning it matches 1 warehouse table
          * "sql": it is based on an SQL statement
        Other types are not supported (and they technically shouldn't be possible.)
        """
        if not table:
            return None

        properties = {
            "id": table.id,
            "name": table.name,
            "type": table.table_type.value,
            "columns": column_properties(table.columns),
        }

        if table.table_type == PhysicalTableType.SQL:
            physical_table = PhysicalTable(
                connection=self.connection,
                id=table.id,
            )
            properties["sql_statement"] = physical_table.sql_statement

        elif table.table_type == PhysicalTableType.NORMAL:
            properties["table_prefix"] = table.table_prefix
            properties["table_name"] = table.table_name

        return properties

    def _logical_table_properties(self, table: LogicalTable) -> dict[str, Any]:
        """
        Returns properties for:
            * the logical table itself
            * its physical table (though it may not be accessible)
            * the columns of the physical table
        """
        physical_table = safe_get_property(table, "physical_table")
        return {
            "id": table.id,
            "name": table.name,
            "physical_table": self._physical_table_properties(physical_table),
        }

    def _report_properties(self, report: Report) -> dict[str, Any]:
        """
        Report properties contain an optional SQL source query. Due to a typing
        bug in the mstrio package, the typing must be ignored.
        """
        properties = self._common_entity_properties(report)  # type: ignore
        properties["url"] = format_url(entity=report, hostname=self.hostname)  # type: ignore
        properties["sql"] = safe_get_property(report, "sql")  # type: ignore
        return properties

    @staticmethod
    def _user_properties(user: User) -> dict[str, Any]:
        return {
            "id": user.id,
            "name": user.name,
            "username": user.username,
            "email": user.default_email_address,
        }

    def _fetch_entities(
        self,
        extract_callback: Callable,
        with_url: bool = True,
        with_description: bool = True,
        custom_property_extractor: Optional[Callable] = None,
    ) -> Iterator[dict[str, Any]]:
        """
        Yields all entities across all projects using the given retrieval function from the mstrio package.
        """
        for project_id in self.project_ids:
            self.connection.select_project(project_id=project_id)

            entities = extract_callback(connection=self.connection)

            for entity in entities:
                try:
                    if custom_property_extractor:
                        yield custom_property_extractor(entity)
                    else:
                        yield self._common_entity_properties(
                            entity,
                            with_url=with_url,
                            with_description=with_description,
                        )
                except IServerError as e:
                    logger.error(
                        f"Could not fetch attributes for entity {entity.id}: {e}"
                    )

    def _fetch_attributes(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(
            list_attributes,
            custom_property_extractor=self._attributes_properties,
        )

    def _fetch_cubes(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(list_olap_cubes)

    def _fetch_dashboards(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(list_dashboards)

    def _fetch_documents(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(list_documents)

    def _fetch_facts(self) -> Iterator[dict[str, Any]]:
        """Yields all facts. Descriptions are not needed for this entity type."""
        return self._fetch_entities(
            list_facts,
            with_url=False,
            with_description=False,
        )

    def _fetch_logical_tables(self) -> Iterator[dict[str, Any]]:
        """
        Yields all logical tables, including their physical tables and their columns.
        """
        return self._fetch_entities(
            list_logical_tables,
            custom_property_extractor=self._logical_table_properties,
        )

    def _fetch_metrics(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(
            list_metrics,
            with_url=False,
        )

    def _fetch_reports(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(
            list_reports,
            custom_property_extractor=self._report_properties,
        )

    def _fetch_users(self) -> Iterator[dict[str, Any]]:
        return self._fetch_entities(
            list_users,
            custom_property_extractor=self._user_properties,
        )

    def fetch(self, asset: StrategyAsset):
        """Fetch the given asset type from Strategy"""
        if asset == StrategyAsset.ATTRIBUTE:
            yield from self._fetch_attributes()

        elif asset == StrategyAsset.CUBE:
            yield from self._fetch_cubes()

        elif asset == StrategyAsset.DASHBOARD:
            yield from self._fetch_dashboards()

        elif asset == StrategyAsset.DOCUMENT:
            yield from self._fetch_documents()

        elif asset == StrategyAsset.FACT:
            yield from self._fetch_facts()

        elif asset == StrategyAsset.LOGICAL_TABLE:
            yield from self._fetch_logical_tables()

        elif asset == StrategyAsset.METRIC:
            yield from self._fetch_metrics()

        elif asset == StrategyAsset.REPORT:
            yield from self._fetch_reports()

        elif asset == StrategyAsset.USER:
            yield from self._fetch_users()

        else:
            raise NotImplementedError(f"Asset type {asset} not implemented yet")



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_SIGMA_"


class SigmaCredentials(BaseSettings):
    """Class to handle Sigma rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    api_token: str = Field(repr=False)
    client_id: str
    host: str
    grant_type: str = "client_credentials"

    @property
    def token_payload(self) -> dict[str, str]:
        return {
            "grant_type": self.grant_type,
            "client_id": self.client_id,
            "client_secret": self.api_token,
        }



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import mode  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-H", "--host", help="Mode Analytics host")
    parser.add_argument("-w", "--workspace", help="Mode Analytics workspace")
    parser.add_argument(
        "-t",
        "--token",
        help="The Token value from the API token",
    )
    parser.add_argument(
        "-s",
        "--secret",
        help="The Password value from the API token",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    mode.extract_all(**parse_filled_arguments(parser))



from collections.abc import Iterator
from typing import Any

from ...utils import group_by
from .constants import SCHEMA_NAME

_HAS_DUPLICATE_KEY = "#has_duplicate_label"


def _clean(raw: str) -> str:
    return raw.strip('"')


def _name(sobject: dict) -> str:
    """
    compute name for table and columns
    - when unique: label
    - when label is empty or has duplicate: label (api_name)
    """
    label = sobject["Label"]
    api_name = sobject["QualifiedApiName"]
    if not label:
        return api_name
    if not sobject[_HAS_DUPLICATE_KEY]:
        return label
    return f"{label} ({api_name})"


def _field_description(field: dict[str, Any]) -> str:
    context: dict[str, str] = {}

    field_definition: dict[str, str] = field.get("FieldDefinition") or {}
    if description := field_definition.get("Description"):
        context["Description"] = _clean(description)
    if help_text := field.get("InlineHelpText"):
        context["Help Text"] = _clean(help_text)
    if compliance_group := field_definition.get("ComplianceGroup"):
        context["Compliance Categorization"] = _clean(compliance_group)
    if security_level := field_definition.get("SecurityClassification"):
        context["Data Sensitivity Level"] = _clean(security_level)

    return "\n".join([f"- {k}: {v}" for k, v in context.items()])


def _to_column_payload(field: dict, position: int, table_name: str) -> dict:
    field_name = _name(field)
    return {
        "column_name": field_name,
        "data_type": field.get("DataType"),
        "description": _field_description(field),
        "id": f"{table_name}.{field_name}",
        "ordinal_position": position,
        "salesforce_developer_name": field.get("DeveloperName"),
        "salesforce_tooling_url": field.get("attributes", {}).get("url"),
        "table_id": table_name,
    }


def _to_table_payload(sobject: dict) -> dict:
    name = _name(sobject)
    return {
        "id": name,
        "api_name": sobject["QualifiedApiName"],
        "label": sobject["Label"],
        "schema_id": SCHEMA_NAME,
        "table_name": name,
        "description": sobject.get("Description"),
        "tags": [],
        "type": "TABLE",
    }


def _remove_duplicates(sobjects: list[dict]) -> list[dict]:
    """only keep one object per QualifiedApiName"""
    by_name = group_by("QualifiedApiName", sobjects)
    return [objects[0] for _, objects in by_name.items()]


def _detect_duplicate_labels(sobjects: list[dict]) -> list[dict]:
    """
    enrich the given data with "has_duplicate_label" flag:
    - True when another asset has the same Label in the list
    - False otherwise
    """
    by_label = group_by("Label", sobjects)
    for sobject in sobjects:
        label = sobject["Label"]
        sobject[_HAS_DUPLICATE_KEY] = len(by_label[label]) > 1
    return sobjects


class SalesforceFormatter:
    """
    Helper functions that format the response in the format to be exported as
    csv.
    """

    @staticmethod
    def tables(sobjects: list[dict]) -> Iterator[dict]:
        """
        formats the raw list of sobjects to tables
        """
        sobjects = _remove_duplicates(sobjects)
        sobjects = _detect_duplicate_labels(sobjects)
        for sobject in sobjects:
            yield _to_table_payload(sobject)

    @staticmethod
    def columns(sobject_fields: dict[str, list[dict]]) -> Iterator[dict]:
        """formats the raw list of sobject fields to columns"""
        for table_name, fields in sobject_fields.items():
            fields = _detect_duplicate_labels(fields)
            for index, field in enumerate(fields):
                yield _to_column_payload(field, index, table_name)



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

METABASE_API_ENV_PREFIX = "CASTOR_METABASE_API_"


class MetabaseApiCredentials(BaseSettings):
    """Metabase's credentials to connect to Metabase API"""

    model_config = SettingsConfigDict(
        env_prefix=METABASE_API_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    base_url: str
    user: str = Field(validation_alias="CASTOR_METABASE_API_USERNAME")
    password: str = Field(repr=False)



from typing import Optional

from pydantic import AliasChoices, ConfigDict, Field
from pydantic.alias_generators import to_camel

from ....utils import PaginationModel

NEXT_PAGE_KEY = "pageToken"


class LookerStudioPagination(PaginationModel):
    items: list = Field(
        default_factory=list,
        validation_alias=AliasChoices("items", "users", "assets"),
    )
    next_page_token: Optional[str] = None

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    def is_last(self) -> bool:
        return self.next_page_token is None

    def next_page_payload(self) -> dict:
        return {NEXT_PAGE_KEY: self.next_page_token}

    def page_results(self) -> list:
        return self.items



import msal  # type: ignore

from ....utils import BearerAuth
from .constants import Keys
from .credentials import PowerbiCredentials
from .endpoints import PowerBiEndpointFactory


class PowerBiBearerAuth(BearerAuth):
    def __init__(self, credentials: PowerbiCredentials):
        self.credentials = credentials
        endpoint_factory = PowerBiEndpointFactory(
            login_url=self.credentials.login_url,
            api_base=self.credentials.api_base,
        )
        authority = endpoint_factory.authority(self.credentials.tenant_id)
        self.app = msal.ConfidentialClientApplication(
            client_id=self.credentials.client_id,
            authority=authority,
            client_credential=self.credentials.secret,
        )

    def fetch_token(self):
        token = self.app.acquire_token_for_client(
            scopes=self.credentials.scopes
        )

        if Keys.ACCESS_TOKEN not in token:
            raise ValueError(f"No access token in token response: {token}")

        return token[Keys.ACCESS_TOKEN]



import pytest

from .batch import batch_of_length


def test_batch_of_length():
    elements = ["a", "b", "c", "d", "e", "f", "g", "h"]
    result = list(batch_of_length(elements, 3))
    assert result == [
        ["a", "b", "c"],
        ["d", "e", "f"],
        ["g", "h"],
    ]

    result = list(batch_of_length(elements, 1000))
    assert result == [
        elements,
    ]

    result = list(batch_of_length(elements, 7))
    assert result == [
        ["a", "b", "c", "d", "e", "f", "g"],
        ["h"],
    ]

    with pytest.raises(AssertionError):
        list(batch_of_length(elements, -12))



from typing import Optional

from pydantic import Field

from ....utils import PaginationModel


class NotionPagination(PaginationModel):
    """Class to handle paginated results"""

    results: list = Field(default_factory=list)
    next_cursor: Optional[str] = None
    has_more: bool = False

    def is_last(self) -> bool:
        return not self.has_more

    def next_page_payload(self) -> dict:
        return {"start_cursor": self.next_cursor}

    def page_results(self) -> list:
        return self.results



import logging
from collections.abc import Iterator
from functools import partial
from typing import Iterable, Optional

import requests
from requests import Response

from ....utils import (
    APIClient,
    BearerAuth,
    RequestSafeMode,
    build_url,
    fetch_all_pages,
    handle_response,
)
from ..assets import (
    ThoughtspotAsset,
)
from .credentials import (
    ThoughtspotCredentials,
)
from .endpoints import (
    ThoughtspotEndpointFactory,
)
from .pagination import METADATA_BATCH_SIZE, ThoughtSpotPagination

_AUTH_TIMEOUT_S = 60
_THOUGHTSPOT_HEADERS = {
    "X-Requested-By": "ThoughtSpot",
    "Accept": "application/json",
    "Content-Type": "application/json",
}
# https://docs.thoughtspot.com/cloud/latest/object-usage-liveboard
_OBJECT_USAGE_LIVEBOARD = "Object Usage"
_ANSWER_USAGE_VIZ = "Answer Usage, by User"
# https://docs.thoughtspot.com/cloud/latest/user-adoption
_USER_ADOPTION_LIVEBOARD = "User Adoption"
_LIVEBOARD_USAGE_VIZ = "Popular Liveboards Last 30 Days"
# By default, no errors are ignored for the moment
THOUGHTSPOT_SAFE_MODE = RequestSafeMode()


logger = logging.getLogger(__name__)


class ThoughtspotBearerAuth(BearerAuth):
    def __init__(self, host: str, token_payload: dict[str, str]):
        auth_endpoint = ThoughtspotEndpointFactory.authentication()
        self.authentication_url = build_url(host, auth_endpoint)
        self.token_payload = token_payload

    def fetch_token(self):
        token_api_path = self.authentication_url
        token_response = requests.post(
            token_api_path, data=self.token_payload, timeout=_AUTH_TIMEOUT_S
        )
        return handle_response(token_response)["token"]


class ThoughtspotClient(APIClient):
    def __init__(
        self,
        credentials: ThoughtspotCredentials,
        safe_mode: Optional[RequestSafeMode] = None,
    ):
        auth = ThoughtspotBearerAuth(
            host=credentials.base_url,
            token_payload=credentials.dict(),
        )
        super().__init__(
            host=credentials.base_url,
            auth=auth,
            headers=_THOUGHTSPOT_HEADERS,
            safe_mode=safe_mode or THOUGHTSPOT_SAFE_MODE,
        )

    def _metadata_search(
        self,
        metadata_type: str,
        identifier: Optional[str] = None,
    ) -> Iterator[dict]:
        """
        Yields assets of the given asset type, and optionally filters on a
        specific identifier.
        """
        offset = 0

        while True:
            search_filters = {
                "metadata": [{"type": metadata_type}],
                "include_details": True,
                "record_size": METADATA_BATCH_SIZE,
                "record_offset": offset,
            }
            if identifier:
                search_filters["metadata"] = {
                    "identifier": identifier,
                    "type": metadata_type,
                }

            metadata = self._post(
                ThoughtspotEndpointFactory.metadata_search(),
                data=search_filters,
            )
            yield from metadata
            if len(metadata) < METADATA_BATCH_SIZE:
                break
            offset = offset + METADATA_BATCH_SIZE

    def _get_all_answers(self) -> Iterator[dict]:
        yield from self._metadata_search(metadata_type="ANSWER")

    def _get_all_liveboards(self) -> Iterator[dict]:
        yield from self._metadata_search(metadata_type="LIVEBOARD")

    def _get_all_columns(self) -> Iterator[dict]:
        yield from self._metadata_search(metadata_type="LOGICAL_COLUMN")

    def _get_all_tables(self) -> Iterator[dict]:
        yield from self._metadata_search(metadata_type="LOGICAL_TABLE")

    def _get_usages(
        self,
        liveboard_name: str,
        visualization_name: str,
    ) -> Iterator[list[list]]:
        """
        Yields the data of a given visualization in the given liveboard.
        ThoughtSpot maintains two system liveboards with stats about data usage,
        which are useful to compute view counts and popularity.
        """
        usage_liveboard = next(
            self._metadata_search(
                metadata_type="LIVEBOARD", identifier=liveboard_name
            )
        )
        liveboard_id = usage_liveboard["metadata_id"]

        def handler(response: Response) -> dict:
            response_dict = response.json()
            contents = response_dict.get("contents", [])
            if not contents:
                logger.warning("No data found in response")
                return dict()
            return contents[0]

        request = partial(
            self._post,
            endpoint=ThoughtspotEndpointFactory.liveboard_data(),
            data={
                "metadata_identifier": liveboard_id,
                "visualization_identifiers": [visualization_name],
                "record_offset": 0,
                "record_size": METADATA_BATCH_SIZE,
            },
            handler=handler,
        )
        yield from fetch_all_pages(request, ThoughtSpotPagination)

    def _get_answer_usages(self) -> Iterator[dict]:
        """
        Returns the usage data of saved Answers, which is found in a visualization
        of the "Object Usage" liveboard.
        Each data row returned by the API is transformed from a list into a dictionary.
        The columns are explicitly listed here because in the API response,
        there is a mismatch between the number of column names and the number
        of values per data row.
        """
        data: Iterable[list[list]] = self._get_usages(
            liveboard_name=_OBJECT_USAGE_LIVEBOARD,
            visualization_name=_ANSWER_USAGE_VIZ,
        )
        columns = (
            "Answer name",
            "Number of unique users",
            "Count of object interactions",
        )
        for row in data:
            yield dict(zip(columns, row))

    def _get_liveboards_usages(self) -> Iterator[dict]:
        """
        Returns the usage data of Liveboards, which is found in a visualization
        of the "User Adoption" liveboard.
        Each data row returned by the API is transformed from a list into a dictionary.
        See `_get_answer_usages` regarding the columns list.
        """
        data: Iterable[list[list]] = self._get_usages(
            liveboard_name=_USER_ADOPTION_LIVEBOARD,
            visualization_name=_LIVEBOARD_USAGE_VIZ,
        )
        columns = ("Pinboard", "Unique Number of User", "Pinboard Views")
        for row in data:
            yield dict(zip(columns, row))

    def fetch(self, asset: ThoughtspotAsset) -> Iterator[dict]:
        if asset == ThoughtspotAsset.ANSWERS:
            yield from self._get_all_answers()

        if asset == ThoughtspotAsset.ANSWER_USAGES:
            yield from self._get_answer_usages()

        if asset == ThoughtspotAsset.LIVEBOARDS:
            yield from self._get_all_liveboards()

        if asset == ThoughtspotAsset.LIVEBOARD_USAGES:
            yield from self._get_liveboards_usages()

        if asset == ThoughtspotAsset.LOGICAL_TABLES:
            yield from self._get_all_tables()



import logging
from collections.abc import Iterable, Iterator
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import NotionAsset
from .client import NotionClient, NotionCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: NotionClient,
) -> Iterable[tuple[NotionAsset, Union[list, Iterator, dict]]]:
    """Iterate over the extracted data from Notion"""

    logger.info("Extracting USERS from API")
    users = list(deep_serialize(client.users()))
    yield NotionAsset.USERS, users
    logger.info(f"Extracted {len(users)} users from API")

    logger.info("Extracting PAGES from API")
    pages = list(deep_serialize(client.pages()))
    yield NotionAsset.PAGES, pages
    logger.info(f"Extracted {len(pages)} pages from API")

    logger.info("Extracting DATABASES from API")
    databases = list(deep_serialize(client.databases()))
    yield NotionAsset.DATABASES, databases
    logger.info(f"Extracted {len(databases)} databases from API")


def extract_all(**kwargs) -> None:
    """
    Extract data from Notion API
    Store the output files locally under the given output_directory
    """

    output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)

    credentials = NotionCredentials(**kwargs)
    client = NotionClient(credentials=credentials)

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), output_directory, ts)
        write_json(filename, data)

    es = current_timestamp()

    write_summary(output_directory, ts, duration_second=es - ts)



COMMON_DASH_LOOK_FIELDS = (
    "id",
    "title",
    "description",
    "folder_id",
    "created_at",
    "deleted_at",
    "last_accessed_at",
    "last_viewed_at",
    "view_count",
    "user_id",
)
FOLDER_FIELDS = (
    "id",
    "name",
    "parent_id",
    "creator_id",
    "is_personal",
    "is_personal_descendant",
)

DASHBOARD_FILTERS = (
    "id",
    "name",
    "title",
    "type",
    "model",
    "name",
    "dimension",
)

DASHBOARD_ELEMENTS = (
    "id",
    "body_text",
    "note_text",
    "subtitle_text",
    "title",
    "title_text",
    "title_hidden",
    "type",
    {
        "query": ("model", "view", "fields"),
        "look": ("id", {"query": ("model", "view", "fields")}),
        "result_maker": (
            "id",
            "vis_config",
            {"query": ("model", "view", "fields")},
        ),
    },
)

DASHBOARD_FIELDS = (
    *COMMON_DASH_LOOK_FIELDS,
    {
        "dashboard_elements": DASHBOARD_ELEMENTS,
        "dashboard_filters": DASHBOARD_FILTERS,
    },
)

LOOK_FIELDS = (
    *COMMON_DASH_LOOK_FIELDS,
    "updated_at",
    "last_updater_id",
    "query_id",
    "is_run_on_load",
    "model",  # { id, label }
)

USER_FIELDS = (
    "id",
    "avatar_url",
    "display_name",
    "email",
    "is_disabled",
    "group_ids",
    "role_ids",
)
LOOKML_FIELDS = (
    "name",
    "label",
    "has_content",
    "project_name",
    {"explores": ("name", "description", "label", "hidden", "group_label")},
)

CONNECTION_FIELDS = (
    "name",
    "dialect",
    "pdts_enabled",
    "host",
    "database",
    "schema",
    "tmp_db_name",
    "jdbc_additional_params",
    "dialect_name",
    "created_at",
    "user_id",
    "example",
    "user_attribute_fields",
    "sql_writing_with_info_schema",
)

PROJECT_FIELDS = (
    "id",
    "name",
    "uses_git",
    "git_remote_url",
    "git_username",
    "git_production_branch_name",
    "use_git_cookie_auth",
    "git_password_user_attribute",
    "git_service_name",
    "git_application_server_http_port",
    "git_application_server_http_scheme",
    "pull_request_mode",
    "validation_required",
    "git_release_mgmt_enabled",
    "allow_warnings",
    "is_example",
)

EXPLORE_FIELD_FIELDS = (
    "description",
    "field_group_label",
    "field_group_variant",
    "hidden",
    "is_filter",
    "label",
    "label_short",
    "measure",
    "name",
    "parameter",
    "primary_key",
    "project_name",
    "scope",
    "sortable",
    "sql",
    "tags",  # string[]
    "type",
    "view",
    "view_label",
    "times_used",
)

EXPLORE_JOIN_FIELDS = (
    "name",
    "dependent_fields",  # string[]
    "fields",  # string[]
    "foreign_key",
    "from_",
    "outer_only",
    "relationship",
    "required_joins",  # string[]
    "sql_foreign_key",
    "sql_on",
    "sql_table_name",
    "type",
    "view_label",
)

EXPLORE_FIELDS = (
    "id",
    "description",
    "name",
    "label",
    "title",
    "scopes",  # string[]
    "project_name",
    "model_name",
    "view_name",
    "hidden",
    "sql_table_name",
    "group_label",
    "tags",  # string[]
    {
        "fields": {
            "dimensions": EXPLORE_FIELD_FIELDS,
            "measures": EXPLORE_FIELD_FIELDS,
        },
        "joins": EXPLORE_JOIN_FIELDS,
    },
)

GROUPS_HIERARCHY_FIELDS = (
    "id",
    "include_by_default",
    "name",
    "user_count",
    "parent_group_ids",
)

GROUPS_ROLES_FIELDS = (
    "can_add_to_content_metadata",
    "id",
    "name",
    "user_count",
    {
        "roles": (
            "id",
            "name",
            {"permission_set": ("id", "name", "permission")},
            {"model_set": ("id", "models", "name")},
        ),
    },
)

CONTENT_VIEWS_FIELDS = (
    "dashboard_id",
    "look_id",
    "start_of_week_date",
    "user_id",
    "view_count",
)
CONTENT_VIEWS_HISTORY_DAYS = 30

USERS_ATTRIBUTES_FIELDS = ("default_value", "id", "label", "name", "type")


# Model from looker
LOOKML_PROJECT_NAME_BLOCKLIST = ("looker-data", "system__activity")



class ExtractionQuery:
    """contains both statement and parameters for the query"""

    def __init__(self, statement: str, params: dict):
        self.statement = statement
        self.params = params



from pytest import raises

from ..utils.url import (
    add_path,
    is_valid,
    url_from,
)


def test_add_path():
    base = "https://test.com"

    # simple
    assert add_path(base, "toto") == f"{base}/toto"

    # multiple parts
    assert add_path(base, "to", "ta") == f"{base}/to/ta"

    # multiple parts with slash
    assert add_path(base, "a/b", "/c/d") == f"{base}/a/b/c/d"

    # base with path
    assert add_path(f"{base}/my/path", "/1/2/", "3") == f"{base}/my/path/1/2/3"

    # base with query string and fragment
    assert add_path(f"{base}?q=2#frag", "1/2") == f"{base}/1/2?q=2#frag"

    # bad base url
    with raises(ValueError):
        add_path("toto", "toto")

    # trailing slash
    base = "https://test.com/"

    # multiple parts with slash
    assert add_path(base, "a/b", "/c/d") == "https://test.com/a/b/c/d"


def test_url_is_valid():
    # valid
    assert is_valid("https://google.com")
    assert is_valid("http://user:pass@test.com:444/my/path?my=query#fragment")
    assert is_valid("ftp://hello.com", valid_schemes=("ftp",))

    # invalid
    assert not is_valid("hello.com")
    assert not is_valid("ftp://hello.com")
    assert not is_valid("http://")


def test_url_from():
    assert url_from() == ""
    assert url_from("http") == "http://"
    assert url_from("https", "google.com") == "https://google.com"
    assert url_from(netloc="te.st", query="q=3") == "//te.st?q=3"



class SigmaEndpointFactory:
    """Wrapper class around all endpoints we're using"""

    DATAMODELS = "dataModels"
    DATASETS = "datasets"
    FILES = "files"
    MEMBERS = "members"
    WORKBOOKS = "workbooks"

    @classmethod
    def authentication(cls) -> str:
        return "v2/auth/token"

    @classmethod
    def connection_path(cls, inode_id: str) -> str:
        return f"v2/connections/paths/{inode_id}"

    @classmethod
    def datamodels(cls) -> str:
        return f"v2/{cls.DATAMODELS}"

    @classmethod
    def datasets(cls) -> str:
        return f"v2/{cls.DATASETS}"

    @classmethod
    def dataset_sources(cls, dataset_id: str) -> str:
        return f"v2/{cls.DATASETS}/{dataset_id}/sources"

    @classmethod
    def elements(cls, workbook_id: str, page_id: str) -> str:
        return f"v2/{cls.WORKBOOKS}/{workbook_id}/pages/{page_id}/elements"

    @classmethod
    def files(cls) -> str:
        return f"v2/{cls.FILES}"

    @classmethod
    def lineage(cls, workbook_id: str, element_id: str) -> str:
        return f"v2/{cls.WORKBOOKS}/{workbook_id}/lineage/elements/{element_id}"

    @classmethod
    def members(cls) -> str:
        return f"v2/{cls.MEMBERS}"

    @classmethod
    def pages(cls, workbook_id: str) -> str:
        return f"v2/{cls.WORKBOOKS}/{workbook_id}/pages"

    @classmethod
    def queries(cls, workbook_id: str) -> str:
        return f"v2/{cls.WORKBOOKS}/{workbook_id}/queries"

    @classmethod
    def workbooks(cls) -> str:
        return f"v2/{cls.WORKBOOKS}"

    @classmethod
    def workbook_sources(cls, workbook_id: str) -> str:
        return f"v2/{cls.WORKBOOKS}/{workbook_id}/sources"



import logging
from collections import defaultdict
from datetime import date

from databricks import sql  # type: ignore

from ...utils import load_file
from .credentials import DatabricksCredentials
from .enums import LineageEntity, TagEntity
from .format import TagMapping
from .utils import build_path, tag_label

logger = logging.getLogger(__name__)

_INFORMATION_SCHEMA_SQL = "SELECT * FROM system.information_schema"

_LINEAGE_SQL_PATHS = {
    LineageEntity.COLUMN: "queries/column_lineage.sql",
    LineageEntity.TABLE: "queries/table_lineage.sql",
}


def _load_lineage_query(lineage_entity: LineageEntity) -> str:
    filename = _LINEAGE_SQL_PATHS[lineage_entity]
    return load_file(filename, __file__)


class DatabricksSQLClient:
    def __init__(
        self,
        credentials: DatabricksCredentials,
        has_table_tags: bool = False,
        has_column_tags: bool = False,
    ):
        self._http_path = credentials.http_path
        self._has_table_tags = has_table_tags
        self._has_column_tags = has_column_tags
        self._host = credentials.host
        self._token = credentials.token

    def execute_sql(self, query: str):
        """
        Execute a SQL query on Databricks system tables and return the results.
        https://docs.databricks.com/en/dev-tools/python-sql-connector.html

        //!\\ credentials.http_path is required in order to run SQL queries
        """
        assert self._http_path, "HTTP_PATH is required to run SQL queries"
        with sql.connect(
            server_hostname=self._host,
            http_path=self._http_path,
            access_token=self._token,
        ) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query)
                return cursor.fetchall()

    def _needs_extraction(self, entity: TagEntity) -> bool:
        if entity == TagEntity.TABLE:
            return self._has_table_tags
        if entity == TagEntity.COLUMN:
            return self._has_column_tags
        raise AssertionError(f"Entity not supported: {entity}")

    def get_tags_mapping(self, entity: TagEntity) -> TagMapping:
        """
        Fetch tags of the given entity and build a mapping:
        { path: list[tags] }

        https://docs.databricks.com/en/sql/language-manual/information-schema/table_tags.html
        https://docs.databricks.com/en/sql/language-manual/information-schema/column_tags.html
        """
        if not self._needs_extraction(entity):
            return dict()

        table = f"{entity.value.lower()}_tags"
        query = f"{_INFORMATION_SCHEMA_SQL}.{table}"
        result = self.execute_sql(query)
        mapping = defaultdict(list)
        for row in result:
            dict_row = row.asDict()
            keys = ["catalog_name", "schema_name", "table_name"]
            if entity == TagEntity.COLUMN:
                keys.append("column_name")
            path = build_path(dict_row, keys)
            label = tag_label(dict_row)
            mapping[path].append(label)

        return mapping

    def get_lineage(
        self,
        lineage_entity: LineageEntity,
        day: date,
    ) -> list[dict]:
        """
        Fetch {TABLE|COLUMN} lineage of the given day, via system tables
        https://docs.databricks.com/en/admin/system-tables/lineage.html

        Unfortunately, passing parameters is not always supported. We have to
        format the query beforehand and pass it as plain text for execution.
        """
        query_template = _load_lineage_query(lineage_entity)
        query = query_template.format(day=day)

        result = self.execute_sql(query)
        data = []
        for row in result:
            data.append(row.asDict())
        return data



import logging
from typing import Optional

import requests
import tableauserverclient as TSC  # type: ignore

from ....utils import SerializedAsset, deduplicate
from ..assets import TableauAsset
from .rest_fields import REST_FIELDS

logger = logging.getLogger(__name__)

_PULSE_API = "api/-/pulse"

_METRICS_DEFINITION_URL = "{base}/pulse/site/{site}/{definition_id}"


def _pick(
    data: SerializedAsset,
    asset: TableauAsset,
) -> SerializedAsset:
    keys = REST_FIELDS[asset]
    return [{key: row[key] for key in keys} for row in data]


class TableauClientRestApi:
    """
    Extract Tableau Assets using REST API
    https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_ref.htm
    """

    def __init__(
        self,
        server: TSC.Server,
    ):
        self._server = server

    @property
    def timeout_sec(self) -> int:
        return self._server.http_options["timeout"]

    @property
    def headers(self) -> dict[str, str]:
        return {"x-tableau-auth": self._server.auth_token}

    def _get_site_name(self) -> str:
        site_id = self._server.site_id
        site = self._server.sites.get_by_id(site_id)
        return site.content_url

    def _get(
        self,
        url: str,
        page_token: Optional[str] = None,
    ) -> dict:
        if page_token:
            url += f"?page_token={page_token}"

        logger.debug(f"Calling REST API: {url}")
        response = requests.get(
            url,
            headers=self.headers,
            timeout=self.timeout_sec,
        )
        response.raise_for_status()
        return response.json()

    def _call(
        self,
        path: str,
        target: str,
    ) -> SerializedAsset:
        base = self._server.server_address.strip("/")
        url = f"{base}/{path}/{target}"

        next_page_token = None
        data = []

        while True:
            response = self._get(url, next_page_token)
            data += response[target]
            next_page_token = response.get("next_page_token")
            if not next_page_token:
                break

        return data

    def _compute_metric_url(self, data: SerializedAsset) -> None:
        site = self._get_site_name()
        base_url = self._server.server_address.strip("/")
        for row in data:
            row["metadata"]["url"] = _METRICS_DEFINITION_URL.format(
                base=base_url,
                site=site,
                definition_id=row["metadata"]["id"],
            )

    def _fetch_metrics(self, definitions: SerializedAsset) -> SerializedAsset:
        metrics = []
        for definition in definitions:
            definition_id = definition["metadata"]["id"]
            path = f"{_PULSE_API}/definitions/{definition_id}"
            metrics += self._call(path=path, target="metrics")

        # for some reason, the REST API sometimes send the same metric twice
        return deduplicate("id", metrics)

    def fetch(
        self,
        asset: TableauAsset,
    ) -> SerializedAsset:
        if asset == TableauAsset.SUBSCRIPTION:
            # https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_ref_pulse.htm#PulseSubscriptionService_ListSubscriptions
            data = self._call(path=_PULSE_API, target="subscriptions")

        elif asset == TableauAsset.METRIC_DEFINITION:
            # https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_ref_pulse.htm#MetricQueryService_ListDefinitions
            data = self._call(path=_PULSE_API, target="definitions")
            self._compute_metric_url(data)

        elif asset == TableauAsset.METRIC:
            # https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_ref_pulse.htm#MetricQueryService_ListMetrics
            definitions = self._call(path=_PULSE_API, target="definitions")
            data = self._fetch_metrics(definitions)
        else:
            raise AssertionError(f"Unsupported asset {asset} for REST API")

        return _pick(data, asset)



from typing import Optional

from pydantic import ConfigDict
from pydantic.alias_generators import to_camel

from ...utils import (
    FetchNextPageBy,
    PaginationModel,
)

LIMIT_RECORDS_PER_PAGE = 2000


class SalesforcePagination(PaginationModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )
    fetch_by: FetchNextPageBy = FetchNextPageBy.URL
    records: list
    next_records_url: Optional[str] = None

    def is_last(self) -> bool:
        no_next_page = not self.next_records_url
        page_incomplete = len(self.records) < LIMIT_RECORDS_PER_PAGE
        return no_next_page or page_incomplete

    def next_page_payload(self) -> Optional[str]:
        return self.next_records_url

    def page_results(self) -> list:
        return self.records



from ..assets import TableauAsset

# list of fields to pick in REST API or TSC responses
REST_FIELDS: dict[TableauAsset, set[str]] = {
    TableauAsset.DATASOURCE: {
        "id",
        "project_id",
        "webpage_url",
    },
    TableauAsset.METRIC: {
        "id",
        "definition_id",
    },
    TableauAsset.METRIC_DEFINITION: {
        "metadata",
        "specification",
    },
    TableauAsset.PROJECT: {
        "description",
        "id",
        "name",
        "parent_id",
    },
    TableauAsset.SUBSCRIPTION: {
        "follower",
        "id",
        "metric_id",
    },
    TableauAsset.USAGE: {
        "name",
        "total_views",
        "workbook_id",
    },
    TableauAsset.USER: {
        "email",
        "fullname",
        "id",
        "name",
        "site_role",
    },
    TableauAsset.WORKBOOK: {
        "id",
        "project_id",
    },
}



from .client import DatabricksClient


class MockDatabricksClient(DatabricksClient):
    def __init__(self):
        self._db_allowed = ["prd", "staging"]
        self._db_blocked = ["dev"]


def test_DatabricksClient__get_user_mapping():
    client = MockDatabricksClient()
    users = [
        {"id": "both", "email": "hello@world.com", "user_name": "hello world"},
        {"id": "no_email", "email": "", "user_name": "no email"},
        {"id": "no_name", "email": "no@name.fr", "user_name": ""},
        {"id": "no_both", "email": "", "user_name": ""},
        {"id": "", "email": "no@id.com", "user_name": "no id"},
    ]
    expected = {
        "hello@world.com": "both",
        "hello world": "both",
        "no@name.fr": "no_name",
        "no email": "no_email",
    }
    mapping = client._get_user_mapping(users)
    assert mapping == expected


def test_DatabricksClient__match_table_with_user():
    client = MockDatabricksClient()
    user_mapping = {"bob@castordoc.com": 3}

    table = {"id": 1, "owner_email": "bob@castordoc.com"}
    table_with_owner = client._match_table_with_user(table, user_mapping)

    assert table_with_owner == {**table, "owner_external_id": 3}

    table_without_owner = {"id": 1, "owner_email": None}
    expected = {"id": 1, "owner_email": None, "owner_external_id": None}
    actual = client._match_table_with_user(table_without_owner, user_mapping)
    assert actual == expected



from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from .....utils import validate_baseurl

QLIK_ENV_PREFIX = "CASTOR_QLIK_"


class QlikCredentials(BaseSettings):
    """
    Qlik's credentials to connect to the API
    """

    model_config = SettingsConfigDict(
        env_prefix=QLIK_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    api_key: str = Field(repr=False)
    base_url: str

    @field_validator("base_url", mode="before")
    @classmethod
    def _check_base_url(cls, base_url: str) -> str:
        return validate_baseurl(base_url)



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_SALESFORCE_"


class SalesforceCredentials(BaseSettings):
    """
    Class to handle Salesforce rest API permissions
    """

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    base_url: str
    client_id: str
    client_secret: str = Field(repr=False)
    password: str = Field(repr=False)
    security_token: str = Field(repr=False)
    username: str

    @property
    def password_token(self) -> str:
        """Generates the password for authentication"""
        return self.password + self.security_token

    def token_request_payload(self) -> dict[str, str]:
        """
        Params to post to the API in order to retrieve the authentication token
        """
        return {
            "grant_type": "password",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": self.username,
            "password": self.password_token,
        }



from typing import Optional, Union

from ....utils import PaginationModel


class CoalescePagination(PaginationModel):
    """
    Class to handle paginated results for Coalesce
    See their documentation here
    https://docs.coalesce.io/docs/api
    """

    data: list
    next: Union[Optional[str], Optional[int]] = None

    def is_last(self) -> bool:
        """Stopping condition for the pagination"""
        return self.next is None

    def next_page_payload(self):
        """Payload enabling to generate the request for the next page"""
        return {"startingFrom": self.next}

    def page_results(self) -> list:
        """List of results of the current page"""
        return self.data



import logging
from typing import Optional

from ..abstract import (
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
    WarehouseAsset,
)

logger = logging.getLogger(__name__)


_NO_DATABASE_ERROR_MSG = (
    "No databases eligible for extraction. "
    "If you are using the db_allow/db_block options, please make sure to use the correct case."
)

_SQLSERVER_DUPLICATED_ASSETS: tuple[WarehouseAsset, ...] = (
    WarehouseAsset.USER,
)


class MSSQLQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from SQL Server.
    """

    def __init__(
        self,
        databases: list[str],
        time_filter: Optional[TimeFilter] = None,
    ):
        super().__init__(
            time_filter=time_filter,
            duplicated=_SQLSERVER_DUPLICATED_ASSETS,
        )
        if not databases:
            raise ValueError(_NO_DATABASE_ERROR_MSG)
        self._databases = databases

    @staticmethod
    def _format(query: ExtractionQuery, values: dict) -> ExtractionQuery:
        return ExtractionQuery(
            statement=query.statement.format(**values),
            params=query.params,
        )

    def build(self, asset: WarehouseAsset) -> list[ExtractionQuery]:
        query = self.build_default(asset)

        if asset == WarehouseAsset.DATABASE:
            # database.sql does not include a {database} placeholder.
            # Indeed, databases are extracted at the server level
            # (not scoped to a specific database).
            return [query]

        logger.info(
            f"\tWill run queries with following database params: {self._databases}",
        )
        return [
            self._format(query, {"database": database})
            for database in self._databases
        ]



import logging
from collections.abc import Iterable, Iterator
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import SigmaAsset
from .client import SigmaClient, SigmaCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: SigmaClient,
) -> Iterable[tuple[SigmaAsset, Union[list, Iterator, dict]]]:
    """Iterate over the extracted data from Sigma"""

    logger.info("Extracting DATA MODELS from API")
    datamodels = client.fetch(SigmaAsset.DATAMODELS)
    yield SigmaAsset.DATASETS, list(deep_serialize(datamodels))

    logger.info("Extracting DATASETS from API")
    datasets = list(client.fetch(SigmaAsset.DATASETS))
    yield SigmaAsset.DATASETS, list(deep_serialize(datasets))

    logger.info("Extracting DATASET SOURCES from API")
    dataset_sources = client.fetch(
        SigmaAsset.DATASET_SOURCES, datasets=datasets
    )
    yield SigmaAsset.DATASET_SOURCES, list(deep_serialize(dataset_sources))

    logger.info("Extracting WORKBOOKS from API")
    workbooks = list(client.fetch(SigmaAsset.WORKBOOKS))
    yield SigmaAsset.WORKBOOKS, list(deep_serialize(workbooks))

    logger.info("Extracting WORKBOOK SOURCES from API")
    workbook_sources = client.fetch(
        SigmaAsset.WORKBOOK_SOURCES, workbooks=workbooks
    )
    yield SigmaAsset.WORKBOOKS, list(deep_serialize(workbook_sources))

    logger.info("Extracting FILES from API")
    files = client.fetch(SigmaAsset.FILES)
    yield SigmaAsset.FILES, list(deep_serialize(files))

    logger.info("Extracting MEMBERS from API")
    members = client.fetch(SigmaAsset.MEMBERS)
    yield SigmaAsset.MEMBERS, list(deep_serialize(members))

    logging.info("Extracting QUERIES data from API")
    queries = client.fetch(SigmaAsset.QUERIES, workbooks=workbooks)
    yield SigmaAsset.QUERIES, list(deep_serialize(queries))

    logging.info("Extracting ELEMENTS data from API")
    elements = list(client.fetch(SigmaAsset.ELEMENTS, workbooks=workbooks))
    yield SigmaAsset.ELEMENTS, list(deep_serialize(elements))

    logging.info("Extracting LINEAGES data from API")
    lineages = client.fetch(SigmaAsset.LINEAGES, elements=elements)
    yield SigmaAsset.LINEAGES, list(deep_serialize(lineages))


def extract_all(**kwargs) -> None:
    """
    Extract data from Sigma API
    Store the output files locally under the given output_directory
    """

    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)

    credentials = SigmaCredentials(**kwargs)
    client = SigmaClient(credentials=credentials)

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts, host=credentials.host)



import logging
from enum import Enum
from typing import Any, Optional

from mstrio.helpers import IServerError  # type: ignore
from mstrio.modeling import (  # type: ignore
    Attribute,
    TableColumn,
)
from mstrio.types import ObjectSubTypes, ObjectTypes  # type: ignore
from mstrio.utils.entity import Entity  # type: ignore
from mstrio.utils.helper import is_dashboard  # type: ignore
from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)

_BATCH_SIZE: int = 100


class URLTemplates(Enum):
    DASHBOARD = "https://{hostname}/MicroStrategyLibrary/app/{project_id}/{id_}"
    DOCUMENT = "https://{hostname}/MicroStrategy/servlet/mstrWeb?documentID={id_}&projectID={project_id}"
    REPORT = "https://{hostname}/MicroStrategy/servlet/mstrWeb?reportID={id_}&projectID={project_id}"
    FOLDER = "https://{hostname}/MicroStrategy/servlet/mstrWeb?folderID={id_}&projectID={project_id}"


class Dependency(BaseModel):
    id: str
    name: str
    subtype: int
    type: int

    model_config = ConfigDict(extra="ignore")


def list_dependencies(entity: Entity) -> list[dict]:
    """Lists the entity's dependencies, keeping only relevant fields."""
    dependencies: list[dict] = []

    offset = 0
    while True:
        batch = entity.list_dependencies(offset=offset, limit=_BATCH_SIZE)
        dependencies.extend(batch)
        if len(batch) < _BATCH_SIZE:
            break
        offset += _BATCH_SIZE

    return [
        Dependency(**dependency).model_dump() for dependency in dependencies
    ]


def _is_dashboard(entity: Entity) -> bool:
    """
    Returns True if the entity is a Dashboard. They can only be distinguished
    from Documents by checking the `view_media` property.
    """
    is_type_document = entity.type == ObjectTypes.DOCUMENT_DEFINITION
    return is_type_document and is_dashboard(entity.view_media)


def _is_report(entity: Entity) -> bool:
    """
    Returns True if the entity is a Report. Cubes share the same type as Reports,
    so the subtype must be checked.
    """
    is_type_report = entity.type == ObjectTypes.REPORT_DEFINITION
    is_subtype_cube = entity.subtype == ObjectSubTypes.OLAP_CUBE.value
    return is_type_report and not is_subtype_cube


def format_url(entity: Entity, hostname: str) -> str:
    """
    Formats the right URL.
    * Dashboards : viewed in MicroStrategy
    * Reports and Documents : viewed in MicroStrategy Web
    * other (i.e. Cubes): the URL leads to the folder in MicroStrategy Web
    """
    if _is_dashboard(entity):
        id_ = entity.id
        template = URLTemplates.DASHBOARD

    elif entity.type == ObjectTypes.DOCUMENT_DEFINITION:
        id_ = entity.id
        template = URLTemplates.DOCUMENT

    elif _is_report(entity):
        id_ = entity.id
        template = URLTemplates.REPORT

    else:
        # default to folder URL
        id_ = level_1_folder_id(entity.ancestors)
        template = URLTemplates.FOLDER

    return template.value.format(
        hostname=hostname,
        id_=id_,
        project_id=entity.project_id,
    )


def safe_get_property(entity: Entity, attribute: str) -> Optional[str]:
    """
    Some properties may raise an error. Example: retrieving a Report's `sql` fails if the Report has not been published.
    This safely returns the attribute value, or None if the retrieval fails.
    """
    try:
        value = getattr(entity, attribute)
    except (IServerError, ValueError) as e:
        logger.error(f"Could not get {attribute} for entity {entity.id}: {e}")
        value = None
    return value


def column_properties(columns: list[TableColumn]) -> list[dict[str, Any]]:
    """Returns the properties of a physical table's columns."""
    properties: list[dict[str, Any]] = []

    for column in columns:
        column_properties = {
            "id": column.id,
            "name": column.name,
            "column_name": column.column_name,
        }
        properties.append(column_properties)

    return properties


def level_1_folder_id(folders: list[dict]) -> str:
    """Searches for the first enclosing folder and returns its ID."""
    for folder in folders:
        if folder["level"] == 1:
            return folder["id"]

    raise ValueError("No level 1 folder found")


def lookup_table_id(attribute: Attribute):
    """Returns the lookup table's ID, if there is one."""
    lookup_table = attribute.attribute_lookup_table
    if not lookup_table:
        return None
    return lookup_table.object_id



from ...types import ExternalAsset, classproperty


class PowerBiAsset(ExternalAsset):
    """PowerBi assets"""

    ACTIVITY_EVENTS = "activity_events"
    DASHBOARDS = "dashboards"
    DATASETS = "datasets"
    DATASET_FIELDS = "dataset_fields"
    DATASET_RELATIONSHIPS = "dataset_relationships"
    METADATA = "metadata"
    PAGES = "pages"
    REPORTS = "reports"
    TABLES = "tables"
    TILES = "tiles"
    USERS = "users"

    @classproperty
    def optional(cls) -> set["PowerBiAsset"]:
        return {
            PowerBiAsset.DATASET_FIELDS,
            PowerBiAsset.DATASET_RELATIONSHIPS,
            PowerBiAsset.PAGES,
            PowerBiAsset.TABLES,
            PowerBiAsset.TILES,
            PowerBiAsset.USERS,
        }



from dataclasses import field

from pydantic_settings import BaseSettings, SettingsConfigDict

DATABRICKS_ENV_PREFIX = "CASTOR_DATABRICKS_"


class DatabricksCredentials(BaseSettings):
    """
    Credentials needed by Databricks client
    Requires:
    - host
    - http_path
    - token
    """

    host: str
    http_path: str
    token: str = field(metadata={"sensitive": True})

    model_config = SettingsConfigDict(
        env_prefix=DATABRICKS_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )



import logging
from typing import Any, Callable, Optional, Union

from requests import HTTPError, Response

logger = logging.getLogger(__name__)

ResponseJson = Union[dict, list[dict]]


class RequestSafeMode:
    """
    RequestSafeMode class to parameterize what should be done if response
    raises due to the status code.

    Attributes:
        self.status_codes: tuple of status codes that will be caught
        self.errors_caught : list of errors caught
    """

    def __init__(
        self,
        max_errors: Union[int, float] = 0,
        status_codes: tuple[int, ...] = (),
    ):
        self.max_errors = max_errors
        self.status_codes: list[int] = list(status_codes)
        self.status_codes_caught: list[int] = []

    def catch_response(self, exception: HTTPError, status_code: int):
        if int(status_code) not in self.status_codes:
            raise exception

        self.status_codes_caught.append(int(status_code))

    @property
    def should_raise(self) -> bool:
        return len(self.status_codes_caught) > self.max_errors


def handle_response(
    response: Response,
    safe_mode: Optional[RequestSafeMode] = None,
    handler: Optional[Callable] = None,
) -> Any:
    """
    Util to handle HTTP Response based on the response status code and the
    safe mode used
    """
    safe_mode = safe_mode if safe_mode else RequestSafeMode()
    try:
        response.raise_for_status()
    except HTTPError as e:
        safe_mode.catch_response(e, response.status_code)
        if safe_mode.should_raise:
            raise e
        logger.error(f"Safe mode : skip request with error {e}")
        logger.debug(e, exc_info=True)
        return {}
    if not handler:
        return response.json()
    return handler(response)



from typing import Optional

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from requests.auth import HTTPBasicAuth

DOMO_ENV_PREFIX = "CASTOR_DOMO_"


class DomoCredentials(BaseSettings):
    """Class to handle Domo rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=DOMO_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    api_token: str = Field(repr=False)
    base_url: str
    client_id: str
    cloud_id: Optional[str] = Field(validation_alias="CLOUD_ID", default=None)
    developer_token: str = Field(repr=False)

    @property
    def authentication(self) -> HTTPBasicAuth:
        return HTTPBasicAuth(self.client_id, self.api_token)

    @property
    def private_headers(self) -> dict[str, str]:
        return {"X-DOMO-Developer-Token": self.developer_token}



import logging
import sys
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from functools import partial

from looker_sdk.error import SDKError
from tqdm import tqdm  # type: ignore

from ...utils import RetryStrategy, deep_serialize, retry
from . import ApiClient
from .assets import LookerAsset

logger = logging.getLogger(__name__)

LOOKER_EXCEPTIONS = (SDKError,)
RETRY_COUNT = 2
RETRY_BASE_MS = 1
RETRY_JITTER_MS = 1
RETRY_STRATEGY = RetryStrategy.LINEAR


@retry(
    exceptions=LOOKER_EXCEPTIONS,
    max_retries=RETRY_COUNT,
    base_ms=RETRY_BASE_MS,
    jitter_ms=RETRY_JITTER_MS,
    strategy=RETRY_STRATEGY,
)
def _make_api_request(
    client: ApiClient,
    asset: LookerAsset,
    folder_id: str,
) -> list:
    """
    Calls the appropriate Looker API endpoint to retrieve either Looks or
    Dashboards withered by the given folder ID.
    """
    if asset == LookerAsset.LOOKS:
        return client.looks(folder_id=folder_id)
    return client.dashboards(folder_id=folder_id)


class MultithreadingFetcher:
    def __init__(
        self,
        folder_ids: set[str],
        client: ApiClient,
        thread_pool_size: int,
        log_to_stdout: bool,
    ):
        self._folder_ids = folder_ids
        self._client = client
        self._thread_pool_size = thread_pool_size
        self._log_to_stdout = log_to_stdout

    def _progress_bar(self, fetch_results: Iterable, total: int) -> tqdm:
        """Create a tqdm progress bar with the appropriate logs destination"""
        file = sys.stderr

        if self._log_to_stdout:
            file = sys.stdout

        return tqdm(fetch_results, total=total, file=file)

    def fetch_assets(self, asset: LookerAsset) -> Iterable[dict]:
        """
        Yields serialized Looks or Dashboards with a request per folder ID.
        Requests are parallelised.
        Explore names associated to dashboards are saved for later use.
        """
        total_assets_count = 0
        total_folders = len(self._folder_ids)

        _fetch = partial(_make_api_request, self._client, asset)

        with ThreadPoolExecutor(max_workers=self._thread_pool_size) as executor:
            fetch_results = executor.map(_fetch, self._folder_ids)

            for results in self._progress_bar(fetch_results, total_folders):
                for result in results:
                    if not result:
                        continue

                    total_assets_count += len(result)
                    yield deep_serialize(result)

        logger.info(f"Fetched {total_assets_count} {asset.value}")



import logging
from collections.abc import Iterator
from typing import Optional

import requests
import tableauserverclient as TSC  # type: ignore

from ....utils import SerializedAsset, retry
from ..assets import TableauAsset
from ..constants import DEFAULT_PAGE_SIZE
from .errors import TableauApiError, TableauApiTimeout
from .gql_queries import FIELDS_QUERIES, GQL_QUERIES, QUERY_TEMPLATE

logger = logging.getLogger(__name__)

# These assets are known to be error-prone, so it's acceptable if a few are missed.
# If errors occur, skip the current batch.
_SAFE_MODE_ASSETS = (
    TableauAsset.COLUMN,
    TableauAsset.FIELD,
)

# increase the value when extraction is too slow
# decrease the value when timeouts arise
_CUSTOM_PAGE_SIZE: dict[TableauAsset, int] = {
    # fields are light but volumes are bigger
    TableauAsset.FIELD: 1000,
    # tables are sometimes heavy
    TableauAsset.TABLE: 50,
}

_TIMEOUT_MESSAGE = (
    "Execution canceled because timeout of 30000 millis was reached"
)

_RETRY_BASE_MS = 10_000
_RETRY_COUNT = 4


def _is_timeout(error: dict) -> bool:
    error_message = error.get("message")
    return error_message == _TIMEOUT_MESSAGE


def _is_warning(error: dict) -> bool:
    extensions = error.get("extensions")
    if not extensions:
        return False

    severity = extensions.get("severity")
    if not severity:
        return False

    return severity.lower() == "warning"


def _check_errors(answer: dict) -> None:
    """
    Handle errors in graphql response:
    - return None when there's no errors in the answer
    - raise TableauApiTimeout if any of the errors is a timeout
    - else raise TableauApiError if any of the errors is critical
    - return None otherwise
    More info about Tableau errors:
    https://help.tableau.com/current/api/metadata_api/en-us/docs/meta_api_errors.html#other-errors
    """
    if "errors" not in answer:
        return

    errors = answer["errors"]

    has_timeout_errors = False
    has_critical_errors = False

    for error in errors:
        if _is_timeout(error):
            has_timeout_errors = True
            continue

        if _is_warning(error):
            # in this case, the answer contains the data anyway
            # just display the warning
            logger.warning(error)
            continue

        # at this point, it's not a timeout error
        # besides, it's not a warning (severity is either "error" or Unknown)
        has_critical_errors = True

    if has_timeout_errors:
        raise TableauApiTimeout(errors)
    if has_critical_errors:
        raise TableauApiError(errors)

    return None


def gql_query_scroll(
    server,
    resource: str,
    fields: str,
    page_size: int,
    skip_batch: bool,
) -> Iterator[SerializedAsset]:
    """
    Iterate over GQL query results, handling pagination and cursor

    We have a retry strategy when timeout issues arise.
    It's a known issue on Tableau side, still waiting for their fix:
    https://issues.salesforce.com/issue/a028c00000zKahoAAC/undefined
    """

    @retry(
        exceptions=(TableauApiTimeout,),
        max_retries=_RETRY_COUNT,
        base_ms=_RETRY_BASE_MS,
    )
    def _call(first: int, offset: int) -> dict:
        query = QUERY_TEMPLATE.format(
            resource=resource,
            fields=fields,
            first=first,
            offset=offset,
        )
        answer = server.metadata.query(query)
        _check_errors(answer)
        return answer["data"][f"{resource}Connection"]

    current_offset = 0
    while True:
        try:
            payload = _call(first=page_size, offset=current_offset)
            yield payload["nodes"]

            current_offset += len(payload["nodes"])
            total = payload["totalCount"]
            logger.info(f"Extracted {current_offset}/{total} {resource}")

            if not payload["pageInfo"]["hasNextPage"]:
                break
        except requests.exceptions.ReadTimeout:
            if not skip_batch:
                raise

            logger.warning("Skipping batch because of TableauServer Timeout")
            current_offset += page_size


def _deduplicate(result_pages: Iterator[SerializedAsset]) -> SerializedAsset:
    """
    Sometimes assets are duplicated, which triggers UniqueViolation errors
    during store_all down the line.

    We suspect the offset pagination to be the root cause, because we had no
    problem until recently, when we switched from cursor pagination to offset
    pagination (for performance reasons)
    https://help.tableau.com/current/api/metadata_api/en-us/docs/meta_api_examples.html#pagination

    This is a straightforward solution to remove these duplicates directly at
    extraction.
    We don't show warnings because duplicates are expected, and we keep only
    the first occurrence since those duplicates are probably identical.
    """
    deduplicated: SerializedAsset = []
    seen_ids: set[str] = set()
    for page in result_pages:
        for asset in page:
            asset_id = asset["id"]
            if asset_id in seen_ids:
                # skip duplicate
                continue
            deduplicated.append(asset)
            seen_ids.add(asset_id)
    return deduplicated


class TableauClientMetadataApi:
    """
    Calls the MetadataAPI, using graphQL
    https://help.tableau.com/current/api/metadata_api/en-us/reference/index.html
    """

    def __init__(
        self,
        server: TSC.Server,
        override_page_size: Optional[int] = None,
    ):
        self._server = server
        self._override_page_size = override_page_size

    def _call(
        self,
        resource: str,
        fields: str,
        page_size: int = DEFAULT_PAGE_SIZE,
        skip_batch: bool = False,
    ) -> SerializedAsset:
        result_pages = gql_query_scroll(
            self._server,
            resource=resource,
            fields=fields,
            page_size=page_size,
            skip_batch=skip_batch,
        )
        return _deduplicate(result_pages)

    def _page_size(self, asset: TableauAsset) -> int:
        return (
            self._override_page_size
            or _CUSTOM_PAGE_SIZE.get(asset)
            or DEFAULT_PAGE_SIZE
        )

    def _fetch_fields(self, skip_batch: bool = False) -> SerializedAsset:
        result: SerializedAsset = []
        page_size = self._page_size(TableauAsset.FIELD)
        for resource, fields in FIELDS_QUERIES:
            current = self._call(
                resource,
                fields,
                page_size,
                skip_batch=skip_batch,
            )
            result.extend(current)
        return result

    @staticmethod
    def _should_skip_batch_with_timeout(
        asset: TableauAsset,
        ignore_metadata_errors: bool = False,
    ) -> bool:
        return asset in _SAFE_MODE_ASSETS and ignore_metadata_errors

    def fetch(
        self,
        asset: TableauAsset,
        ignore_errors: bool = False,
    ) -> SerializedAsset:
        skip_batch = self._should_skip_batch_with_timeout(asset, ignore_errors)

        if asset == TableauAsset.FIELD:
            return self._fetch_fields(skip_batch=skip_batch)

        page_size = self._page_size(asset)
        resource, fields = GQL_QUERIES[asset]
        return self._call(
            resource=resource,
            fields=fields,
            page_size=page_size,
            skip_batch=skip_batch,
        )



import logging
from collections.abc import Iterator

from sqlalchemy import text

from ...utils import ExtractionQuery, SqlalchemyClient, uri_encode

logger = logging.getLogger(__name__)

SERVER_URI = "{user}:{password}@{host}:{port}"
MSSQL_URI = f"mssql+pymssql://{SERVER_URI}"
DEFAULT_PORT = 1433

_KEYS = ("user", "password", "host", "port")

_SYSTEM_DATABASES = (
    "DBAdmin",
    "dbaTools",
    "master",
    "model",
    "msdb",
    "tempdb",
)


def _check_key(credentials: dict) -> None:
    for key in _KEYS:
        if key not in credentials:
            raise KeyError(f"Missing {key} in credentials")


class MSSQLClient(SqlalchemyClient):
    """Microsoft Server SQL client"""

    @staticmethod
    def name() -> str:
        return "MSSQL"

    def _engine_options(self, credentials: dict) -> dict:
        return {}

    def _build_uri(self, credentials: dict) -> str:
        _check_key(credentials)
        uri = MSSQL_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            host=credentials["host"],
            port=credentials.get("port") or DEFAULT_PORT,
        )
        return uri

    def execute(self, query: ExtractionQuery) -> Iterator[dict]:
        """
        Re-implements the SQLAlchemyClient execute function to ensure we consume
        the cursor before calling connection.close() as it wipes out the data
        otherwise
        """
        connection = self.connect()
        try:
            proxy = connection.execute(text(query.statement), query.params)
            results = list(self._process_result(proxy))
            yield from results
        finally:
            self.close()

    def get_databases(self) -> list[str]:
        result = self.execute(
            ExtractionQuery("SELECT name FROM sys.databases", {})
        )
        return [
            row["name"]
            for row in result
            if row["name"] not in _SYSTEM_DATABASES
        ]

    def _has_access(self, name: str, object_type: str, permission: str) -> bool:
        query_text = f"""
            SELECT
            HAS_PERMS_BY_NAME('{name}', '{object_type}', '{permission}')
            AS has_permission
        """
        query = ExtractionQuery(query_text, dict())
        result = next(self.execute(query))
        return result["has_permission"] == 1

    def _has_table_read_access(self, database: str, table_name: str) -> bool:
        """
        Check whether we have READ access to the given table
        """
        return self._has_access(
            name=f"[{database}].{table_name}",
            object_type="OBJECT",
            permission="SELECT",
        )

    def _has_view_database_state(self, database: str) -> bool:
        """
        Check whether we have VIEW DATABASE STATE permissions, which
        is necessary to fetch data from the Query Store
        """
        return self._has_access(
            name=database,
            object_type="DATABASE",
            permission="VIEW DATABASE STATE",
        )

    def _has_query_store(self, database: str) -> bool:
        """
        Checks whether the Query Store is activated on this database.
        This is required to extract the SQL queries history.
        https://learn.microsoft.com/en-us/sql/relational-databases/performance/monitoring-performance-by-using-the-query-store?view=sql-server-ver17"""
        sql = f"""
        SELECT
            desired_state
        FROM
            [{database}].sys.database_query_store_options
        """
        query = ExtractionQuery(sql, {})
        # 2 = READ_WRITE, which means the Query Store is activated
        return next(self.execute(query))["desired_state"] == 2

    def has_queries_permissions(self, database: str) -> bool:
        """
        Verify that we have the required permissions to extract
        query history and view object definitions (DDL).

        This check ensures:
        - Query Store is enabled on the database.
        - We have the VIEW DATABASE STATE permissions
        - We have read access to the relevant system tables.
        """

        tables = (
            # SQL queries
            "sys.query_store_plan",
            "sys.query_store_query",
            "sys.query_store_query_text",
            "sys.query_store_runtime_stats",
            # views DDL
            "sys.schemas",
            "sys.sql_modules",
            "sys.views",
        )

        has_permissions = True
        for table in tables:
            if not self._has_table_read_access(database, table):
                logger.info(f"Missing READ permission on {database}.{table}")
                has_permissions = False

        if not self._has_view_database_state(database):
            logger.info(f"Missing VIEW DATABASE STATE in database {database}")
            has_permissions = False

        if not self._has_query_store(database):
            logger.info(f"Query Store is not activated in database {database}")
            has_permissions = False

        return has_permissions



import logging
from collections.abc import Iterator
from functools import partial
from typing import Iterable, Optional

from ....utils import (
    APIClient,
    BasicAuth,
    fetch_all_pages,
)
from ..assets import (
    ConfluenceAsset,
)
from .credentials import ConfluenceCredentials
from .endpoints import ConfluenceEndpointFactory
from .pagination import ConfluencePagination

logger = logging.getLogger(__name__)

_HEADERS = {
    "Accept": "application/json",
    "Content-Type": "application/json",
}
_STATUS_ARCHIVED = "archived"
_TYPE_PERSONAL = "personal"


class ConfluenceClient(APIClient):
    def __init__(
        self,
        credentials: ConfluenceCredentials,
        include_archived_spaces: bool = False,
        include_personal_spaces: bool = False,
        space_ids_allowed: Optional[set[str]] = None,
        space_ids_blocked: Optional[set[str]] = None,
    ):
        self.account_id = credentials.account_id
        auth = BasicAuth(
            username=credentials.username, password=credentials.token
        )
        super().__init__(
            auth=auth,
            host=credentials.base_url,
            headers=_HEADERS,
        )

        self.include_archived_spaces = include_archived_spaces
        self.include_personal_spaces = include_personal_spaces
        self.space_ids_allowed = space_ids_allowed or set()
        self.space_ids_blocked = space_ids_blocked or set()

    def databases(self, database_ids: Iterable[str]) -> Iterator[dict]:
        """
        Extracts all given databases
        """
        for _id in database_ids:
            database = self._get(
                endpoint=ConfluenceEndpointFactory.database(_id),
            )
            yield database

    def folders(self, folder_ids: Iterable[str]) -> Iterator[dict]:
        """
        Extracts all given folders and their parent folders.
        """
        candidate_ids = set(folder_ids)
        seen = set()

        while candidate_ids:
            folder_id = candidate_ids.pop()
            if folder_id in seen:
                continue

            seen.add(folder_id)
            folder = self._get(
                endpoint=ConfluenceEndpointFactory.folder(folder_id),
            )
            yield folder

            parent_type = folder.get("parentType")
            if parent_type == "folder":
                folder_id = folder["parentId"]
                if folder_id not in seen:
                    candidate_ids.add(folder["parentId"])

            if not parent_type:
                logger.info(f"folder with unknown parent: {folder_id}")

    def pages(self):
        """Extracts all pages from all relevant Spaces."""
        for space in self.spaces():
            space_id = space["id"]
            request = partial(
                self._get,
                endpoint=ConfluenceEndpointFactory.pages(space_id),
            )
            yield from fetch_all_pages(request, ConfluencePagination)

    def spaces(self) -> Iterator[dict]:
        """
        Returns the spaces meeting the conditions defined by the settings.

        If `space_ids_allowed` is not empty, only matching spaces are returned.

        Otherwise, all spaces are filtered by excluding the following:
          * The space is in the blocked list
          * The space is personal (type=personal) and skip_personal_spaces is True
          * The space is archived (status=archived) and skip_archived_spaces is True
        """
        request = partial(
            self._get,
            endpoint=ConfluenceEndpointFactory.spaces(),
        )
        spaces = list(fetch_all_pages(request, ConfluencePagination))

        if self.space_ids_allowed:
            yield from (
                space
                for space in spaces
                if space["id"] in self.space_ids_allowed
            )
            return

        for space in spaces:
            space_id = space["id"]
            type_ = space["type"]
            status = space["status"]

            if space_id in self.space_ids_blocked:
                continue

            if status == _STATUS_ARCHIVED and not self.include_archived_spaces:
                continue

            if type_ == _TYPE_PERSONAL and not self.include_personal_spaces:
                continue

            yield space

    def users(self):
        request_body = {"accountIds": [self.account_id]}
        request = partial(
            self._post,
            endpoint=ConfluenceEndpointFactory.users(),
            data=request_body,
        )
        yield from fetch_all_pages(request, ConfluencePagination)

    def fetch(
        self,
        asset: ConfluenceAsset,
        *,
        folder_ids: Optional[Iterator[str]] = None,
        database_ids: Optional[Iterator[str]] = None,
    ) -> Iterator[dict]:
        """Returns the needed metadata for the queried asset"""
        if asset == ConfluenceAsset.FOLDERS:
            assert folder_ids is not None
            yield from self.folders(folder_ids)

        elif asset == ConfluenceAsset.DATABASES:
            assert database_ids is not None
            yield from self.databases(database_ids)

        elif asset == ConfluenceAsset.PAGES:
            yield from self.pages()

        elif asset == ConfluenceAsset.USERS:
            yield from self.users()

        else:
            raise ValueError(f"This asset {asset} is unknown")



import logging
from collections.abc import Iterator
from functools import partial
from http import HTTPStatus
from typing import Optional

import requests
from requests import HTTPError, Response

from ...utils import (
    APIClient,
    BearerAuth,
    build_url,
    fetch_all_pages,
    handle_response,
)
from .constants import DEFAULT_API_VERSION, DEFAULT_PAGINATION_LIMIT
from .credentials import SalesforceCredentials
from .pagination import SalesforcePagination

logger = logging.getLogger(__name__)

SALESFORCE_TIMEOUT_S = 120


class SalesforceBadRequestError(HTTPError):
    """
    Custom Exception to print the response's text when an error occurs
    during Salesforce's authentication.
    """

    def __init__(self, response: Response):
        text = response.text
        message = (
            f"{response.status_code} Client Error: {response.reason} for url: {response.url}"
            f"\nResponse text: {text}"
        )
        super().__init__(message, response=response)


class SalesforceAuth(BearerAuth):
    _AUTH_ENDPOINT = "services/oauth2/token"

    def __init__(self, credentials: SalesforceCredentials):
        self._host = credentials.base_url
        self._token_payload = credentials.token_request_payload()

    def fetch_token(self) -> Optional[str]:
        """
        Fetches the access token from Salesforce using the provided credentials.
        A custom Exception is raised if the request fails with a 400 status code.
        """
        url = build_url(self._host, self._AUTH_ENDPOINT)
        response = requests.post(url, "POST", params=self._token_payload)

        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise SalesforceBadRequestError(response)

        handled_response = handle_response(response)
        return handled_response["access_token"]


class SalesforceBaseClient(APIClient):
    """
    Salesforce API client.
    https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/intro_rest.htm
    """

    api_version = DEFAULT_API_VERSION
    pagination_limit = DEFAULT_PAGINATION_LIMIT

    PATH_TPL = "services/data/v{version}/{suffix}"

    def __init__(self, credentials: SalesforceCredentials):
        auth = SalesforceAuth(credentials)
        super().__init__(
            host=credentials.base_url, auth=auth, timeout=SALESFORCE_TIMEOUT_S
        )

    def _endpoint(self, suffix: str) -> str:
        path = self.PATH_TPL.format(version=self.api_version, suffix=suffix)
        return path

    @property
    def query_endpoint(self) -> str:
        """Returns the query API url"""
        return self._endpoint("query")

    @property
    def tooling_endpoint(self) -> str:
        """Returns the tooling API url"""
        return self._endpoint("tooling/query")

    def _query_all(self, query: str) -> Iterator[dict]:
        """
        Run a SOQL query over salesforce API.

        more: https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm
        """
        request = partial(
            self._get,
            endpoint=self.query_endpoint,
            params={"q": query},
        )
        yield from fetch_all_pages(request, SalesforcePagination)



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_COALESCE_"


class CoalesceCredentials(BaseSettings):
    """Class to handle Coalesce rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    host: str
    token: str = Field(repr=False)

    @property
    def token_payload(self) -> dict[str, str]:
        return {
            "client_secret": self.token,
        }



import logging
from typing import Optional

from ...utils import mapping_from_rows
from ..abstract import TimeFilter
from .api_client import DatabricksAPIClient
from .credentials import DatabricksCredentials
from .enums import TagEntity
from .format import DatabricksFormatter
from .sql_client import DatabricksSQLClient
from .types import TablesColumns

logger = logging.getLogger(__name__)

_THREADS_COLUMN_LINEAGE = 2
_THREADS_TABLE_LINEAGE = 10


class DatabricksClient:
    """Databricks Client"""

    def __init__(
        self,
        credentials: DatabricksCredentials,
        db_allowed: Optional[set[str]] = None,
        db_blocked: Optional[set[str]] = None,
        has_table_tags: bool = False,
        has_column_tags: bool = False,
    ):
        self.api_client = DatabricksAPIClient(
            credentials=credentials,
            db_allowed=db_allowed,
            db_blocked=db_blocked,
        )
        self.sql_client = DatabricksSQLClient(
            credentials=credentials,
            has_table_tags=has_table_tags,
            has_column_tags=has_column_tags,
        )

        self.formatter = DatabricksFormatter()

    @staticmethod
    def name() -> str:
        return "Databricks"

    @staticmethod
    def _match_table_with_user(table: dict, user_mapping: dict) -> dict:
        """Matches the table's owner email to an ID, or None if not found."""
        table_owner_email = table.get("owner_email")
        owner_external_id = (
            user_mapping.get(table_owner_email) if table_owner_email else None
        )
        return {**table, "owner_external_id": owner_external_id}

    @staticmethod
    def _get_user_mapping(users: list[dict]) -> dict:
        return {
            **mapping_from_rows(users, "email", "id"),
            **mapping_from_rows(users, "user_name", "id"),
        }

    def schemas(self, databases: list[dict]) -> list[dict]:
        return self.api_client.schemas(databases)

    def databases(self) -> list[dict]:
        return self.api_client.databases()

    def tables_and_columns(
        self, schemas: list[dict], users: list[dict]
    ) -> TablesColumns:
        """
        Get the databricks tables & columns leveraging the unity catalog API
        """
        tables: list[dict] = []
        columns: list[dict] = []
        user_mapping = self._get_user_mapping(users)
        table_tags = self.sql_client.get_tags_mapping(TagEntity.TABLE)
        column_tags = self.sql_client.get_tags_mapping(TagEntity.COLUMN)
        for schema in schemas:
            t_to_add, c_to_add = self.api_client.tables_columns_of_schema(
                schema=schema,
                table_tags=table_tags,
                column_tags=column_tags,
            )
            t_with_owner = [
                self._match_table_with_user(table, user_mapping)
                for table in t_to_add
            ]
            tables.extend(t_with_owner)
            columns.extend(c_to_add)
        return tables, columns

    def queries(self, time_filter: Optional[TimeFilter] = None) -> list[dict]:
        return self.api_client.queries(time_filter)

    def users(self) -> list[dict]:
        return self.api_client.users()

    def view_ddl(self, schemas: list[dict]) -> list[dict]:
        return self.api_client.view_ddl(schemas)



from ...types import ExternalAsset, classproperty


class LookerAsset(ExternalAsset):
    """Looker assets"""

    CONNECTIONS = "connections"
    CONTENT_VIEWS = "content_views"
    DASHBOARDS = "dashboards"
    EXPLORES = "explores"
    FOLDERS = "folders"
    GROUPS_HIERARCHY = "groups_hierarchy"
    GROUPS_ROLES = "groups_roles"
    LOOKML_MODELS = "lookml_models"
    LOOKS = "looks"
    PROJECTS = "projects"
    USERS = "users"
    USERS_ATTRIBUTES = "users_attributes"

    @classproperty
    def optional(cls) -> set["LookerAsset"]:
        return {
            LookerAsset.CONNECTIONS,
            LookerAsset.CONTENT_VIEWS,
            LookerAsset.GROUPS_HIERARCHY,
            LookerAsset.GROUPS_ROLES,
            LookerAsset.PROJECTS,
            LookerAsset.USERS_ATTRIBUTES,
        }



from typing import Optional

from pydantic import UUID4, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from .constant import FileType
from .enums import Zone

UPLOADER_ENV_PREFIX = "CASTOR_UPLOADER_"


class UploaderSettings(BaseSettings):
    """Class holding Castor uploader attributes"""

    model_config = SettingsConfigDict(
        env_prefix=UPLOADER_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    directory_path: Optional[str] = None
    file_path: Optional[str] = None
    file_type: FileType
    source_id: UUID4
    token: str = Field(repr=False)
    zone: Optional[Zone] = Zone.EU
